# Generated by Django 3.2.13 on 2023-01-20 16:46

import ddm.models.core
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0029_rename_eventlog_eventlogentry'),
    ]

    operations = [
        migrations.AddField(
            model_name='donationproject',
            name='img_header_left',
            field=models.ImageField(blank=True, null=True, upload_to=ddm.models.core.project_header_dir_path),
        ),
        migrations.AddField(
            model_name='donationproject',
            name='img_header_right',
            field=models.ImageField(blank=True, null=True, upload_to=ddm.models.core.project_header_dir_path),
        ),
    ]



# Generated by Django 3.2.13 on 2023-01-04 19:26

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0023_auto_20230104_1325'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomToken',
            fields=[
                ('key', models.CharField(max_length=40, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('expiration_date', models.DateTimeField(blank=True, null=True)),
                ('project', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='donation_project', to='ddm.donationproject', verbose_name='Donation Project')),
            ],
        ),
    ]



import datetime

from django.test import TestCase, override_settings
from django.urls import reverse
from django.utils import timezone
from django.contrib.auth import get_user_model

from ddm.models.auth import ProjectTokenAuthenticator, ProjectAccessToken
from ddm.models.core import ResearchProfile, DonationProject
from ddm.auth import email_is_valid, user_is_owner, user_is_permitted

from rest_framework import exceptions


User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestAuthenticationFlow(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.base_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        }
        cls.base_user = User.objects.create_user(**cls.base_creds)
        cls.base_profile = ResearchProfile.objects.create(user=cls.base_user)

        cls.not_permitted_creds = {
            'username': 'no_per', 'password': '123', 'email': 'noperm@liam.com'
        }
        cls.not_permitted_user = User.objects.create_user(**cls.not_permitted_creds)
        cls.not_permitted_profile = ResearchProfile.objects.create(user=cls.not_permitted_user)

        cls.wo_profile_creds = {
            'username': 'no_prof', 'password': '123', 'email': 'noprof@mail.com'
        }
        cls.wo_profile_user = User.objects.create_user(**cls.wo_profile_creds)

    def test_email_is_valid(self):
        valid_email = 'abc@mail.com'
        invalid_email = 'abc@liam.com'
        self.assertTrue(email_is_valid(valid_email))
        self.assertFalse(email_is_valid(invalid_email))

    def test_user_is_permitted_without_permission(self):
        self.assertFalse(user_is_permitted(self.not_permitted_user))

    def test_user_is_permitted_with_ignore_email_restriction_true(self):
        self.not_permitted_profile.ignore_email_restriction = True
        self.not_permitted_profile.save()
        self.assertTrue(user_is_permitted(self.not_permitted_user))
        self.not_permitted_profile.ignore_email_restriction = False
        self.not_permitted_profile.save()

    def test_user_is_owner(self):
        project = DonationProject.objects.create(name='test-project', slug='test',
                                                 owner=self.base_profile)
        self.assertTrue(user_is_owner(self.base_user, project.pk))
        self.assertFalse(user_is_owner(self.wo_profile_user, project.pk))

    def test_ddm_no_permission_view_200(self):
        self.client.login(**self.not_permitted_creds)
        response = self.client.get(reverse('ddm-no-permission'), follow=True)
        self.assertEqual(response.status_code, 200)

    def test_ddm_no_permission_view_redirect_to_project_list(self):
        self.client.login(**self.base_creds)
        response = self.client.get(reverse('ddm-no-permission'), follow=True)
        self.assertRedirects(response, reverse('project-list'))


class TestAuthCustomTokenAuthenticator(TestCase):

    @classmethod
    def setUpTestData(cls):
        # User
        base_creds = {
            'username': 'base_user', 'password': '123', 'email': 'base@mail.com'
        }
        base_user = User.objects.create_user(**base_creds)
        base_user_profile = ResearchProfile.objects.create(user=base_user)

        # Project
        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=base_user_profile
        )

        # Authenticator
        cls.authenticator = ProjectTokenAuthenticator()
        cls.token = ProjectAccessToken.objects.create(
            project=cls.project, created=timezone.now(), expiration_date=None
        )

    def test_valid_token_without_expiration(self):
        validated_token = self.authenticator.authenticate_credentials(
            self.token, self.project.pk)[1]
        self.assertEqual(self.token, validated_token)

    def test_valid_token_with_expiration(self):
        self.token.delete()
        token = ProjectAccessToken.objects.create(
            project=self.project,
            created=timezone.now(),
            expiration_date=timezone.now() + datetime.timedelta(days=2)
        )
        validated_token = self.authenticator.authenticate_credentials(
            token, self.project.pk)[1]
        self.assertEqual(token, validated_token)

    def test_invalid_token(self):
        token = 'rubbish'
        self.assertRaises(
            exceptions.AuthenticationFailed,
            self.authenticator.authenticate_credentials,
            token, self.project.pk
        )

    def test_without_token(self):
        self.assertRaises(
            exceptions.AuthenticationFailed,
            self.authenticator.authenticate_credentials,
            None, self.project.pk
        )

    def test_expired_token(self):
        self.token.delete()
        expired_token = ProjectAccessToken.objects.create(
            project=self.project,
            created=timezone.now(),
            expiration_date=datetime.datetime(2022, 2, 2, 22, 22).replace(tzinfo=timezone.utc)
        )
        self.assertRaises(
            exceptions.AuthenticationFailed,
            self.authenticator.authenticate_credentials,
            expired_token, self.project.pk
        )



from django.contrib import admin
from django.urls import reverse
from django.utils.html import format_html
from ddm.models.core import DonationProject
from ddm.models.logs import ExceptionLogEntry, EventLogEntry


class DonationProjectAdmin(admin.ModelAdmin):
    """
    Provides an overview of all active Donation Projects.
    """
    list_display = ['name', 'owner', 'date_created', 'edit_link']
    readonly_fields = ['date_created']

    def has_add_permission(self, request, obj=None):
        return False

    @admin.display(description="Link to Project")
    def edit_link(self, obj):
        url = reverse('project-detail', args=[obj.pk])
        return format_html(f'<a href="{url}">Show Detail Page</a>')


class ExceptionsAdmin(admin.ModelAdmin):
    """
    Provides an overview of all registered exceptions.
    """
    list_display = [
        'formatted_date', 'project', 'raised_by', 'exception_type', 'message'
    ]

    def has_add_permission(self, request, obj=None):
        return False

    @admin.display(ordering='date', description='Date')
    def formatted_date(self, obj):
        return obj.date.strftime('%Y-%m-%d %H:%M:%S')


class EventsAdmin(admin.ModelAdmin):
    """
    Provides an overview of all registered exceptions.
    """
    list_display = ['formatted_date', 'project', 'description', 'message']
    list_filter = ['description']

    def has_add_permission(self, request, obj=None):
        return False

    @admin.display(ordering='date', description='Date')
    def formatted_date(self, obj):
        return obj.date.strftime('%Y-%m-%d %H:%M:%S')


admin.site.register(DonationProject, DonationProjectAdmin)
admin.site.register(ExceptionLogEntry, ExceptionsAdmin)
admin.site.register(EventLogEntry, EventsAdmin)



# Generated by Django 3.2.9 on 2022-03-17 15:10

import ckeditor.fields
import ddm.models.questions
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='DonationBlueprint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=250)),
                ('exp_file_format', models.CharField(choices=[('json', 'Json Format')], default='json', max_length=10)),
                ('expected_fields', models.JSONField()),
                ('extracted_fields', models.JSONField()),
                ('regex_path', models.TextField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='DonationProject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30)),
                ('slug', models.SlugField(unique=True, verbose_name='External Project Slug')),
                ('intro_text', ckeditor.fields.RichTextField(blank=True, null=True)),
                ('outro_text', ckeditor.fields.RichTextField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Participant',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start_time', models.DateTimeField()),
                ('end_time', models.DateTimeField(null=True)),
                ('completed', models.BooleanField(default=False)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
            ],
        ),
        migrations.CreateModel(
            name='QuestionBase',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('question_type', models.CharField(choices=[('generic', 'Generic Question'), ('single_choice', 'Single Choice Question'), ('multi_choice', 'Multi Choice Question'), ('matrix', 'Matrix Question'), ('semantic_diff', 'Semantic Differential'), ('open', 'Open Question'), ('transition', 'Transition Block')], default='generic', max_length=20)),
                ('name', models.CharField(max_length=255)),
                ('index', models.PositiveIntegerField(default=1)),
                ('variable_name', models.SlugField()),
                ('text', ckeditor.fields.RichTextField(blank=True, null=True)),
                ('required', models.BooleanField(default=False)),
                ('blueprint', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationblueprint')),
                ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_ddm.questionbase_set+', to='contenttypes.contenttype')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
            ],
            options={
                'ordering': ['index'],
            },
        ),
        migrations.CreateModel(
            name='MatrixQuestion',
            fields=[
                ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm.questionbase')),
                ('randomize_items', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
            bases=(ddm.models.questions.ScaleMixin, 'ddm.questionbase', models.Model),
        ),
        migrations.CreateModel(
            name='MultiChoiceQuestion',
            fields=[
                ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm.questionbase')),
                ('randomize_items', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
            bases=('ddm.questionbase', models.Model),
        ),
        migrations.CreateModel(
            name='OpenQuestion',
            fields=[
                ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm.questionbase')),
                ('display', models.CharField(choices=[('small', 'Small'), ('large', 'Large')], default='large', max_length=20)),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
            },
            bases=('ddm.questionbase',),
        ),
        migrations.CreateModel(
            name='SemanticDifferential',
            fields=[
                ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm.questionbase')),
                ('randomize_items', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
            bases=(ddm.models.questions.ScaleMixin, 'ddm.questionbase', models.Model),
        ),
        migrations.CreateModel(
            name='SingleChoiceQuestion',
            fields=[
                ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm.questionbase')),
                ('randomize_items', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
            bases=('ddm.questionbase', models.Model),
        ),
        migrations.CreateModel(
            name='Transition',
            fields=[
                ('questionbase_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='ddm.questionbase')),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
            },
            bases=('ddm.questionbase',),
        ),
        migrations.CreateModel(
            name='ZippedBlueprint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=250)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
            ],
        ),
        migrations.CreateModel(
            name='ScalePoint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('index', models.IntegerField()),
                ('label', models.CharField(blank=True, max_length=100, null=True)),
                ('value', models.IntegerField()),
                ('add_border', models.BooleanField(default=False)),
                ('question', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.questionbase')),
            ],
            options={
                'ordering': ['index'],
            },
        ),
        migrations.CreateModel(
            name='QuestionnaireResponse',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('time_submitted', models.DateTimeField()),
                ('data', models.JSONField()),
                ('participant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.participant')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
            ],
        ),
        migrations.CreateModel(
            name='QuestionItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.CharField(blank=True, max_length=255, null=True)),
                ('label_alt', models.CharField(blank=True, max_length=255, null=True)),
                ('index', models.IntegerField()),
                ('value', models.IntegerField()),
                ('randomize', models.BooleanField(default=False)),
                ('question', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.questionbase')),
            ],
            options={
                'ordering': ['index'],
            },
        ),
        migrations.CreateModel(
            name='DonationInstruction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', ckeditor.fields.RichTextField(blank=True, null=True)),
                ('index', models.PositiveIntegerField(default=1)),
                ('blueprint', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='ddm.donationblueprint')),
                ('zip_blueprint', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='ddm.zippedblueprint')),
            ],
            options={
                'ordering': ['index'],
            },
        ),
        migrations.AddField(
            model_name='donationblueprint',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject'),
        ),
        migrations.AddField(
            model_name='donationblueprint',
            name='zip_blueprint',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='ddm.zippedblueprint'),
        ),
        migrations.CreateModel(
            name='DataDonation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('time', models.DateTimeField()),
                ('consent', models.BooleanField(default=False)),
                ('status', models.JSONField()),
                ('data', models.JSONField()),
                ('blueprint', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='ddm.donationblueprint')),
                ('participant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.participant')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
            ],
        ),
        migrations.AddConstraint(
            model_name='scalepoint',
            constraint=models.UniqueConstraint(fields=('index', 'question'), name='unique_index_per_question'),
        ),
        migrations.AddConstraint(
            model_name='questionbase',
            constraint=models.UniqueConstraint(fields=('variable_name', 'project'), name='unique_varname_per_project'),
        ),
        migrations.AddConstraint(
            model_name='donationinstruction',
            constraint=models.UniqueConstraint(fields=('index', 'blueprint'), name='unique_index_per_blueprint'),
        ),
        migrations.AddConstraint(
            model_name='donationinstruction',
            constraint=models.UniqueConstraint(fields=('index', 'zip_blueprint'), name='unique_index_per_zipblueprint'),
        ),
    ]



import io
import json
import zipfile

from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse
from django.views.decorators.debug import sensitive_variables

from ddm.models.auth import ProjectTokenAuthenticator
from ddm.models.core import (
    DataDonation, DonationProject, QuestionnaireResponse, ResearchProfile,
    Participant, DonationBlueprint
)
from ddm.models.encryption import Decryption
from ddm.models.logs import EventLogEntry, ExceptionLogEntry
from ddm.models.serializers import (
    DonationSerializer, ResponseSerializer, ProjectSerializer, ParticipantSerializer
)

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import authentication, exceptions, permissions, status


def user_is_allowed(user, project):
    """
    Returns true if user is owner or collaborator of a project. False otherwise.
    """
    if user.is_anonymous:
        return False

    user_profile = ResearchProfile.objects.filter(user=user).first()
    if not user_profile:
        return False
    else:
        if project.owner != user_profile:
            return False
        else:
            return True


class DDMAPIMixin:
    """
    Mixin containing ddm-specific methods to be combined with DRF-views.
    """
    def permission_denied(self, request, message=None, code=None):
        """
        If request is not permitted, determine what kind of exception to raise.
        Added EventLog entries.
        """
        if request.authenticators and not request.successful_authenticator:
            self.create_event_log(
                descr='Failed Attempt',
                msg='Authentication failed.'
            )
            raise exceptions.NotAuthenticated()

        self.create_event_log(
            descr='Failed Attempt',
            msg='Permission Denied.'
        )
        raise exceptions.PermissionDenied(detail=message, code=code)

    def get_project(self):
        """ Returns project instance. """
        return DonationProject.objects.filter(pk=self.kwargs['pk']).first()

    def create_event_log(self, descr, msg):
        """ Creates an event log entry related to the current project. """
        if self.request is not None:
            prefix = f'{self.request.get_full_path()} {self.request.method}: '
        else:
            prefix = ''
        return EventLogEntry.objects.create(project=self.get_project(),
                                            description=f'{prefix}{descr}',
                                            message=msg)


class ProjectDataAPI(APIView, DDMAPIMixin):
    """
    Download all data collected for a given donation project.

    Returns:
    - GET: A Response object with the complete data associated to a project (i.e.,
    donated data, questionnaire responses, metadata) and status code.

    Example Usage:
    ```
    GET /api/project/<project_pk>/data

    Returns a ZIP-Folder containing a json file with the following structure:
    {
        'project': {<project information>},
        'donations': {<collected donations per file blueprint>},
        'questionnaire': {<questionnaire responses>}
        'participants': {<participant information>}
    }
    ```

    Authentication Methods:
    - Token authentication for remote calls.
    - Session authentication for access through web application (by verifying
        that the requesting user is the project owner).

    Error Responses:
    - 400 Bad Request: If there's an issue with the input data.
    - 401 Unauthorized: If authentication fails.
    - 403 Forbidden: If a user is not permitted to access a project (session
        authentication only).
    """
    authentication_classes = [ProjectTokenAuthenticator,
                              authentication.SessionAuthentication]
    permission_classes = [permissions.IsAuthenticated]

    @sensitive_variables()
    def get(self, request, format=None, *args, **kwargs):
        """
        Return a zip container that contains a json file which holds the
        data donations and questionnaire responses.
        """
        project = self.get_project()
        if not user_is_allowed(request.user, project):
            self.create_event_log(
                descr='Forbidden Download Request',
                msg='Request user is not permitted to download the data.'
            )
            msg = 'User does not have access.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        # Extract secret from request if project is super secret.
        secret = project.secret_key
        if project.super_secret:
            super_secret = None if 'Super-Secret' not in request.headers else request.headers['Super-Secret']
            if super_secret is None:
                self.create_event_log(
                    descr='Failed Download Attempt',
                    msg='Download requested without supplying secret.'
                )
                msg = 'Incorrect key material.'
                return Response(status=status.HTTP_403_FORBIDDEN,
                                data={'message': msg})
            else:
                secret = super_secret

        # Gather project data in dictionary.
        blueprints = DonationBlueprint.objects.filter(project=project)
        q_responses = QuestionnaireResponse.objects.filter(project=project)
        participants = Participant.objects.filter(project=project)
        try:
            decryptor = Decryption(secret, project.get_salt())

            donations = {}
            for blueprint in blueprints:
                blueprint_donations = blueprint.datadonation_set.all()
                donations[blueprint.name] = [DonationSerializer(d, decryptor=decryptor).data for d in blueprint_donations]

            results = {
                'project': ProjectSerializer(project).data,
                'donations': donations,
                'questionnaire': [ResponseSerializer(r, decryptor=decryptor).data for r in q_responses],
                'participants': [ParticipantSerializer(p).data for p in participants]
            }
        except ValueError:
            self.create_event_log(
                descr='Failed Download Attempt',
                msg='Download requested with incorrect secret.'
            )
            msg = 'Incorrect key material.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        # Create zip file.
        zip_in_mem = self.create_zip(results)
        response = self.create_zip_response(zip_in_mem)
        self.create_event_log(
            descr='Data Download Successful',
            msg='The project data was downloaded.'
        )
        return response

    @staticmethod
    def create_zip(content):
        """ Creates a zip file in memory. """
        buffer = io.BytesIO()
        with zipfile.ZipFile(buffer, mode='w', compression=zipfile.ZIP_DEFLATED) as zf:
            with zf.open('data.json', 'w') as json_file:
                json_file.write(json.dumps(content, ensure_ascii=False, separators=(',', ':')).encode('utf-8'))
                zf.testzip()
        zip_in_memory = buffer.getvalue()
        buffer.flush()
        return zip_in_memory

    @staticmethod
    def create_zip_response(zip_file):
        """ Creates an HttpResponse object containing the provided zip file. """
        response = HttpResponse(zip_file, content_type='application/zip')
        response['Content-Length'] = len(zip_file)
        response['Content-Disposition'] = 'attachment; filename=zipfile.zip'
        return response


class DeleteProjectData(APIView, DDMAPIMixin):
    """
    Delete all data (i.e., data donations and questionnaire responses)
    collected for a given donation project.

    Returns:
    - A Response object with the status code.

    Example Usage:
    ```
    DELETE /api/project/<project_pk>/data/delete
    {
        'status': '200',
        'data': {'message': 'Deleted <n donations> data donations and
                            <n responses> questionnaire responses.'}
    }
    ```

    Authentication Methods:
    - Session authentication for access through web application (by verifying
        that the requesting user is the project owner).

    Error Responses:
    - 400 Bad Request: If there's an issue with the input data.
    - 401 Unauthorized: If authentication fails.
    - 403 Forbidden: If a user is not permitted to access a project.
    """
    authentication_classes = [authentication.SessionAuthentication]
    permission_classes = [permissions.IsAuthenticated]

    def delete(self, request, format=None, *args, **kwargs):
        """
        Delete data donations and questionnaire requests associated with project.
        """
        project = self.get_project()
        if not user_is_allowed(request.user, project):
            self.create_event_log(
                descr='Forbidden Deletion Request',
                msg='Request user is not permitted to delete the project data.'
            )
            msg = 'User does not have access.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        # Delete all related objects.
        Participant.objects.filter(project=project).delete()
        n_deleted_donations = DataDonation.objects.filter(project=project).delete()[0]
        n_deleted_responses = QuestionnaireResponse.objects.filter(project=project).delete()[0]

        self.create_event_log(
            descr='Data Deletion Successful',
            msg='The project data was deleted.'
        )

        msg = (f'Deleted {n_deleted_donations} data donations and '
               f'{n_deleted_responses} questionnaire responses.')
        return Response(status=status.HTTP_200_OK, data={'message': msg})


class DeleteParticipantAPI(APIView, DDMAPIMixin):
    """
    Delete a participant of an owned project by providing their external ID.

    * Session authentication for browser access.
    * Only project owners are able to access this view.

    Returns:
    - A Response object with the status code.

    Authentication Methods:
    - Token authentication for remote calls.
    - Session authentication for access through web application (by verifying
        that the requesting user is the project owner).

    Error Responses:
    - 400 Bad Request: If there's an issue with the input data.
    - 401 Unauthorized: If authentication fails.
    - 403 Forbidden: If a user is not permitted to access a project.
    """
    authentication_classes = [ProjectTokenAuthenticator,
                              authentication.SessionAuthentication]
    permission_classes = [permissions.IsAuthenticated]

    def delete(self, request, format=None, *args, **kwargs):
        """
        Delete participant related to current project by providing external_id.
        """
        project = self.get_project()
        if not user_is_allowed(request.user, project):
            self.create_event_log(
                descr='Forbidden Deletion Request',
                msg='Request user is not permitted to delete the project data.'
            )
            msg = 'User does not have access.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        external_id = self.kwargs['participant_id']
        try:
            participant = Participant.objects.get(project=project, external_id=external_id)
        except ObjectDoesNotExist:
            msg = f'Participant with id "{external_id}" not found.'
            return Response(status=status.HTTP_404_NOT_FOUND, data={'message': msg})

        participant.delete()

        self.create_event_log(
            descr=f'Participant with external id "{external_id}" successfully deleted.',
            msg='Participant deleted.'
        )

        msg = f'Participant with external id "{external_id}" successfully deleted.'
        return Response(status=status.HTTP_200_OK, data={'message': msg})


class ExceptionAPI(APIView):

    permission_classes = [permissions.AllowAny]

    def post(self, request, format=None, *args, **kwargs):
        """
        Log exception messages received from client.
        """
        project_id = self.kwargs['pk']
        project = DonationProject.objects.get(pk=project_id)

        try:
            participant_id = request.session[f'project-{project_id}']['participant_id']
            participant = Participant.objects.get(pk=participant_id)
        except KeyError:
            participant = None

        blueprint_id = request.data.get('blueprint')
        if blueprint_id:
            blueprint = DonationBlueprint.objects.get(pk=blueprint_id)
        else:
            blueprint = None

        ExceptionLogEntry.objects.create(
            project=project,
            participant=participant,
            exception_type=request.data.get('status_code'),
            message=request.data['message'],
            raised_by=request.data.get('raised_by'),
            blueprint=blueprint
        )

        return Response(None, status=201)


class DonationsAPI(APIView, DDMAPIMixin):
    """
    Retrieve a set of Donations collected for a given Donation Project.

    This can be useful to create feedback dashboards for participants. For example,
    after completing a study, you could redirect your participants to another
    website where you retrieve the data donated by the participant and display
    some personalized insights. This can be useful to incentivize participation.

    Returns:
    - A Response object with data and status code.

    Available query filters:
    - participants: A comma separated list of external participant IDs
        belonging to the project whose donations should be returned
        (e.g. ?participants='ID1,ID2,ID3').
        If not specified, donations for all participants are returned.
    - blueprints: A comma separated list of file blueprint IDs, for which the
        donations should be returned (e.g. ?blueprints='BP-ID1,BP-ID2,BP-ID3').
        If not specified, donations for all blueprints associated with the
        project are returned.

    Example Usage:
    ```
    GET /api/project/<project_pk>/donations
    {
        "BP-Name": [[<donations participant A>], [<donations participant B>], ...],
        "BP2-Name": ...
    }
    ```

    Error Responses:
    - 400 Bad Request: If there's an issue with the input data.
    - 401 Unauthorized: If authentication fails.
    - 405 Method Not Allowed: If resources belonging to a super secret project
        are requested.

    Note:
    No identifying information is included in the returned object - i.e., no
    participant IDs are included. This means that if donations for multiple
    participants are requested, donations cannot be linked to specific
    participants.
    Please also note that this endpoint cannot be used for secret projects.
    """
    authentication_classes = [ProjectTokenAuthenticator]
    permission_classes = [permissions.IsAuthenticated]

    def get_blueprints(self, project):
        blueprints = self.request.query_params.get('blueprints')
        if blueprints is not None:
            return DonationBlueprint.objects.filter(
                project=project,
                pk__in=blueprints.split(',')
            )
        else:
            return DonationBlueprint.objects.filter(project=project)

    def get_donations(self, blueprint):
        participants = self.request.query_params.get('participants')
        if participants is not None:
            return blueprint.datadonation_set.filter(
                participant__external_id__in=participants.split(','))
        else:
            return blueprint.datadonation_set.all()

    @sensitive_variables()
    def get(self, request, format=None, *args, **kwargs):
        project = self.get_project()

        if project.super_secret:
            msg = 'Endpoint disabled for super secret projects.'
            return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED,
                            data={'message': msg})

        blueprints = self.get_blueprints(project)
        try:
            decryptor = Decryption(project.secret_key, project.get_salt())
            donations = {}
            for blueprint in blueprints:
                blueprint_donations = self.get_donations(blueprint)
                donations[blueprint.name] = [DonationSerializer(d, decryptor=decryptor).data['data']
                                             for d in blueprint_donations]

        except ValueError:
            self.create_event_log(
                descr='Failed Attempt',
                msg='Download requested with incorrect secret.'
            )
            msg = 'Incorrect key material.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        response = Response(donations)
        self.create_event_log(
            descr='Data Download Successful',
            msg='The project data was downloaded.'
        )
        return response


class ResponsesAPI(APIView, DDMAPIMixin):
    """
    Retrieve a set of Questionnaire Responses collected for a given Donation Project.

    This can be useful to create feedback dashboards for participants. For example,
    after completing a study, you could redirect your participants to another
    website where you retrieve the data provided by the participant and display
    some personalized insights. This can be useful to incentivize participation.

    Returns:
    - A Response object with data and status code.

    Available query filters:
    - participants: A comma separated list of external participant IDs
        belonging to the project whose donations should be returned
        (e.g. ?participants='ID1,ID2,ID3').
        If not specified, donations for all participants are returned.

    Example Usage:
    ```
    GET /api/project/<project_pk>/responses
    [{'var_name_q1': 'answer', 'varname_q2': 'answer', ... },
     {<responses participant B>}, ...]
    ```

    Error Responses:
    - 400 Bad Request: If there's an issue with the input data.
    - 401 Unauthorized: If authentication fails.
    - 405 Method Not Allowed: If resources belonging to a super secret project
        are requested.

    Note:
    No identifying information is included in the returned object - i.e., no
    participant IDs are included. This means that if responses for multiple
    participants are requested, responses cannot be linked to specific
    participants.
    Please also note that this endpoint cannot be used for secret projects.
    """
    authentication_classes = [ProjectTokenAuthenticator]
    permission_classes = [permissions.IsAuthenticated]

    def get_responses(self, project):
        participants = self.request.query_params.get('participants')
        if participants is not None:
            return QuestionnaireResponse.objects.filter(
                project=project, participant__external_id__in=participants.split(','))
        else:
            return QuestionnaireResponse.objects.filter(project=project)

    @sensitive_variables()
    def get(self, request, format=None, *args, **kwargs):
        project = self.get_project()

        if project.super_secret:
            msg = 'Endpoint disabled for super secret projects.'
            return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED,
                            data={'message': msg})

        try:
            decryptor = Decryption(project.secret_key, project.get_salt())
            responses = [ResponseSerializer(r, decryptor=decryptor).data['responses']
                         for r in self.get_responses(project)]

        except ValueError:
            self.create_event_log(
                descr='Failed Attempt',
                msg='Download requested with incorrect secret.'
            )
            msg = 'Incorrect key material.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        response = Response(responses)
        self.create_event_log(
            descr='Data Download Successful',
            msg='The project data was downloaded.'
        )
        return response



# Generated by Django 3.2.13 on 2024-03-13 15:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0046_donationblueprint_expected_fields_regex_matching'),
    ]

    operations = [
        migrations.AddField(
            model_name='fileuploader',
            name='combined_consent',
            field=models.BooleanField(default=False, help_text='If enabled, participants will be asked to consent to submit all uploaded data at once. Otherwise, participant will be asked to consent to the submission of the data separately for each blueprint.', verbose_name='All-in-one consent'),
        ),
    ]



from django.contrib.auth import get_user_model
from django.http import HttpResponseRedirect, Http404
from django.shortcuts import reverse, redirect
from django.urls import reverse_lazy
from django.views.generic import CreateView
from django.views.generic.base import TemplateView

from ddm.auth import user_is_permitted, user_is_owner
from ddm.forms import ResearchProfileConfirmationForm
from ddm.models.core import ResearchProfile


User = get_user_model()


class DdmAuthMixin:
    """
    Mixin for Class Based Views that handles redirects as follows:
    - unauthenticated users => login page.
    - users without research profiles => create research profile.
    - users without authorization (e.g., due to e-mail restriction) => no permission page.
    - users without owner-rights => 404.
    """
    def dispatch(self, request, *args, **kwargs):
        if request.method == 'GET':
            if not request.user.is_authenticated:
                return redirect('ddm-login')
            elif not user_is_permitted(request.user):
                return redirect('ddm-no-permission')
            elif not ResearchProfile.objects.filter(user=request.user).exists():
                ResearchProfile.objects.create(user=request.user)
            else:
                if request.path not in [reverse('project-list'), reverse('project-create')]:
                    if 'project_pk' in self.kwargs:
                        project_pk = self.kwargs['project_pk']
                    else:
                        project_pk = self.kwargs['pk']
                    if not user_is_owner(request.user, project_pk) and not request.user.is_superuser:
                        raise Http404()
        return super().dispatch(request, *args, **kwargs)


class DdmRegisterResearchProfileView(CreateView):
    """
    View to create a research profile for a user.
    Implements the following redirects:
    * Logged-in users with a research profile are redirected to the project list.
    * Unauthenticated users are redirected to the login page.
    * Users without permission rights are redirected to the no-permission view.
    """
    model = ResearchProfile
    form_class = ResearchProfileConfirmationForm
    template_name = 'ddm/admin/auth/confirm_profile.html'
    success_url = reverse_lazy('project-list')

    def dispatch(self, request, *args, **kwargs):
        if request.method == 'GET':
            if not request.user.is_authenticated:
                return redirect('ddm-login')
            elif not user_is_permitted(request.user):
                return redirect('ddm-no-permission')
            elif ResearchProfile.objects.filter(user=request.user).exists():
                return redirect('project-list')
        return super().dispatch(request, *args, **kwargs)

    def get_initial(self):
        self.initial = super().get_initial()
        self.initial.update({'user': self.request.user.pk})
        return self.initial

    def form_invalid(self, form):
        return HttpResponseRedirect(reverse('ddm-no-permission'))


class DdmNoPermissionView(TemplateView):
    """
    View to inform users that they do not have the needed permission rights.
    Implements the following redirects:
    * Unauthenticated users are redirected to the login page.
    * Logged-in users with permission and a research profile are redirected to the project list.
    """
    template_name = 'ddm/admin/auth/no_permission.html'

    def dispatch(self, request, *args, **kwargs):
        if request.method == 'GET':
            if not request.user.is_authenticated:
                return redirect('ddm-login')
            elif (user_is_permitted(request.user) and
                  ResearchProfile.objects.filter(user=request.user).exists()):
                return redirect('project-list')
        return super().dispatch(request, *args, **kwargs)



# Generated by Django 3.2.13 on 2022-09-22 14:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0013_auto_20220922_1324'),
    ]

    operations = [
        migrations.AddField(
            model_name='donationproject',
            name='redirect_enabled',
            field=models.BooleanField(default=False, verbose_name='Redirect enabled'),
        ),
        migrations.AddField(
            model_name='donationproject',
            name='redirect_target',
            field=models.CharField(blank=True, max_length=2000, null=True, verbose_name='Redirect target'),
        ),
    ]



# Generated by Django 3.2.13 on 2023-02-17 14:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0034_auto_20230210_1148'),
    ]

    operations = [
        migrations.AddField(
            model_name='processingrule',
            name='replacement_value',
            field=models.TextField(blank=True),
        ),
        migrations.AlterField(
            model_name='processingrule',
            name='comparison_operator',
            field=models.CharField(blank=True, choices=[('==', 'Equal (==)'), ('!=', 'Not Equal (!=)'), ('>', 'Greater than (>)'), ('<', 'Smaller than (<)'), ('>=', 'Greater than or equal (>=)'), ('<=', 'Smaller than or equal (<=)'), ('regex-delete-match', 'Delete match (regex)'), ('regex-replace-match', 'Replace match (regex)'), ('regex-delete-row', 'Delete row when match (regex)')], default=None, max_length=24, null=True),
        ),
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase

from ddm.forms import ProjectCreateForm
from ddm.models.core import ResearchProfile


User = get_user_model()


class TestProjectCreateForm(TestCase):
    @classmethod
    def setUpTestData(cls):
        user_credentials = {
            'username': 'user', 'password': '123', 'email': 'base@mail.com'
        }
        user = User.objects.create_user(**user_credentials)
        cls.user_profile = ResearchProfile.objects.create(user=user)

    def test_form_raises_error_if_super_secret_without_secret(self):
        post_data = {
            'name': 'project',
            'slug': 'project',
            'super_secret': True,
            'secret': '',
            'owner': self.user_profile
        }
        form = ProjectCreateForm(post_data)
        self.assertFalse(form.is_valid())



from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.db.models import Q
from django.http import Http404, HttpResponseRedirect
from django.urls import reverse
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.views.generic.list import ListView

from ddm.forms import BlueprintEditForm, ProcessingRuleInlineFormset
from ddm.models.core import DonationBlueprint, DonationInstruction, DonationProject, FileUploader
from ddm.views.admin import DdmAuthMixin


class BlueprintMixin:
    """ Mixin for all blueprint related views. """
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        file_uploaders = FileUploader.objects.filter(project__pk=self.kwargs['project_pk'])
        context.update({
            'project': DonationProject.objects.get(pk=self.kwargs['project_pk']),
            'file_uploader_meta': {str(fu.pk): fu.upload_type for fu in file_uploaders}
        })
        return context

    def get_success_url(self):
        return reverse('data-donation-overview', kwargs={'project_pk': self.kwargs['project_pk']})


class DataDonationOverview(DdmAuthMixin, BlueprintMixin, ListView):
    """ View to list all file uploaders associated with a project. """
    model = FileUploader
    context_object_name = 'file_uploaders'
    template_name = 'ddm/admin/data_donation/overview.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'lonely_blueprints': context['project'].donationblueprint_set.filter(file_uploader=None)})
        return context

    def get_queryset(self):
        queryset = super().get_queryset().filter(project_id=self.kwargs['project_pk'])
        return queryset


class FileUploaderCreate(SuccessMessageMixin, DdmAuthMixin, BlueprintMixin, CreateView):
    """ View to create a new file uploader. """
    model = FileUploader
    template_name = 'ddm/admin/data_donation/file_uploader/create.html'
    fields = ['name', 'upload_type', 'combined_consent']
    success_message = 'File Uploader was created successfully.'

    def form_valid(self, form):
        form.instance.project_id = self.kwargs['project_pk']
        return super().form_valid(form)

    def get_success_url(self):
        kwargs = {
            'project_pk': self.kwargs['project_pk'],
            'pk': self.object.pk
        }
        return reverse('file-uploader-edit', kwargs=kwargs)


class FileUploaderEdit(SuccessMessageMixin, DdmAuthMixin, BlueprintMixin, UpdateView):
    """ View to edit the details of an existing file uploader. """
    model = FileUploader
    template_name = 'ddm/admin/data_donation/file_uploader/edit.html'
    fields = ['name', 'upload_type', 'combined_consent', 'index']
    success_message = 'Blueprint Uploader "%(name)s" was successfully updated.'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'blueprints': self.get_relevant_blueprints()})
        return context

    def get_relevant_blueprints(self):
        """
        Returns a query set containing all blueprints that are associated with
        the current or no file uploader.
        """
        relevant_blueprints = DonationBlueprint.objects.filter(
            Q(file_uploader=self.object) | Q(file_uploader=None), project__pk=self.kwargs['project_pk'])
        return relevant_blueprints

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        form_class = self.get_form_class()
        form = self.get_form(form_class)

        # Add custom error.
        selected_blueprints = [int(k[3:]) for k in self.request.POST.keys() if k.startswith('bp-')]

        # Validate form.
        if form.is_valid():
            return self.form_valid(form, selected_blueprints)
        else:
            return self.form_invalid(form)

    def form_valid(self, form, selected_blueprints):
        self.object = form.save()

        # Update file_upload foreign key on donation blueprints.
        relevant_blueprints = self.get_relevant_blueprints()
        for bp in relevant_blueprints:
            if bp.pk in selected_blueprints:
                bp.file_uploader = self.object
                bp.save()
            elif bp.file_uploader == self.object:
                bp.file_uploader = None
                bp.save()
        messages.add_message(
            self.request, messages.SUCCESS,
            self.success_message % dict(name=self.object.name),
            fail_silently=True,
        )
        return HttpResponseRedirect(self.get_success_url())


class FileUploaderDelete(DdmAuthMixin, BlueprintMixin, DeleteView):
    """ View to delete an existing blueprint uploader. """
    model = FileUploader
    template_name = 'ddm/admin/data_donation/file_uploader/delete.html'
    success_message = 'File Uploader "%s" was deleted.'

    def delete(self, request, *args, **kwargs):
        messages.success(self.request, self.success_message % self.get_object().name)
        return super().delete(request, *args, **kwargs)


class BlueprintCreate(SuccessMessageMixin, DdmAuthMixin, BlueprintMixin, CreateView):
    """ View to create a new donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm/admin/data_donation/donation_blueprint/create.html'
    form_class = BlueprintEditForm
    success_message = 'Blueprint was created successfully.'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        available_file_uploaders = FileUploader.objects.filter(project_id=self.kwargs['project_pk'])
        context['form'].fields['file_uploader'].queryset = available_file_uploaders
        return context

    def form_valid(self, form):
        form.instance.project_id = self.kwargs['project_pk']
        return super().form_valid(form)

    def get_success_url(self):
        return reverse('blueprint-edit', kwargs={'project_pk': self.object.project.pk, 'pk': self.object.pk})


class BlueprintEdit(SuccessMessageMixin, DdmAuthMixin, BlueprintMixin, UpdateView):
    """ View to edit the details of an existing donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm/admin/data_donation/donation_blueprint/edit.html'
    form_class = BlueprintEditForm
    success_message = 'Blueprint "%(name)s" was successfully updated.'

    def get_success_url(self):
        return reverse('data-donation-overview', kwargs={'project_pk': self.object.project.pk})

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        available_file_uploaders = FileUploader.objects.filter(project_id=self.kwargs['project_pk'])
        context['form'].fields['file_uploader'].queryset = available_file_uploaders
        context['formset'] = ProcessingRuleInlineFormset(
            instance=self.object, queryset=self.object.processingrule_set.order_by('execution_order'))
        return context

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        form_class = self.get_form_class()
        form = self.get_form(form_class)
        formset = ProcessingRuleInlineFormset(self.request.POST, instance=self.object)
        if form.is_valid() and formset.is_valid():
            return self.form_valid(form, formset)
        else:
            return self.form_invalid(form, formset)

    def form_valid(self, form, formset):
        self.object = form.save()
        formset.instance = self.object
        formset.save()
        messages.add_message(
            self.request, messages.SUCCESS,
            self.success_message % dict(name=self.object.name),
            fail_silently=True,
        )
        return HttpResponseRedirect(self.get_success_url())

    def form_invalid(self, form, formset):
        return self.render_to_response(self.get_context_data(form=form, formset=formset))


class BlueprintDelete(DdmAuthMixin, BlueprintMixin, DeleteView):
    """ View to delete an existing donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm/admin/data_donation/donation_blueprint/delete.html'
    success_message = 'Blueprint "%s" was deleted.'

    def delete(self, request, *args, **kwargs):
        messages.success(self.request, self.success_message % self.get_object().name)
        return super().delete(request, *args, **kwargs)


class InstructionMixin:
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'project_pk': self.kwargs['project_pk']})
        context.update({'file_uploader': FileUploader.objects.get(pk=self.kwargs['file_uploader_pk'])})
        return context

    def get_success_url(self):
        kwargs = {'project_pk': self.kwargs['project_pk'],
                  'file_uploader_pk': self.kwargs['file_uploader_pk']}
        return reverse('instruction-overview', kwargs=kwargs)


class InstructionOverview(DdmAuthMixin, InstructionMixin, ListView):
    """ View to create a new instruction page. """
    model = DonationInstruction
    context_object_name = 'instructions'
    template_name = 'ddm/admin/data_donation/instructions/list.html'
    fields = ['text', 'index']

    def get_queryset(self):
        queryset = super().get_queryset().filter(file_uploader_id=self.kwargs['file_uploader_pk'])
        return queryset


class InstructionCreate(SuccessMessageMixin, DdmAuthMixin, InstructionMixin, CreateView):
    """ View to create an instruction page. """
    model = DonationInstruction
    template_name = 'ddm/admin/data_donation/instructions/create.html'
    fields = ['text', 'index']
    success_message = 'Instruction page was successfully created.'

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['instance'] = DonationInstruction(file_uploader=FileUploader.objects.get(id=self.kwargs['file_uploader_pk']))
        return kwargs

    def get_initial(self):
        initial = super().get_initial()
        related_file_uploader = FileUploader.objects.get(id=self.kwargs['file_uploader_pk'])
        indices = related_file_uploader.donationinstruction_set.values_list('index', flat=True)
        if indices:
            initial['index'] = max(indices) + 1
        else:
            initial['index'] = 1
        return initial


class InstructionEdit(SuccessMessageMixin, DdmAuthMixin, InstructionMixin, UpdateView):
    """ View to edit an instruction page. """
    model = DonationInstruction
    template_name = 'ddm/admin/data_donation/instructions/edit.html'
    fields = ['text', 'index']
    success_message = 'Instruction page was successfully updated.'


class InstructionDelete(DdmAuthMixin, InstructionMixin, DeleteView):
    """ View to delete an instruction page. """
    model = DonationInstruction
    template_name = 'ddm/admin/data_donation/instructions/delete.html'
    success_message = 'Instruction page was deleted.'

    def delete(self, request, *args, **kwargs):
        messages.success(self.request, self.success_message)
        return super().delete(request, *args, **kwargs)



import cProfile
import inspect
import json
import pandas as pd

from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from django.urls import reverse
from django.utils import timezone

from ddm.models.core import (
    ResearchProfile, DonationProject, FileUploader, DonationBlueprint,
    Participant, DataDonation
)


User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class PerformanceTest(TestCase):
    """ Defines shared test data. """
    @classmethod
    def setUpTestData(cls):
        cls.user_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        }
        user = User.objects.create_user(**cls.user_creds)
        profile = ResearchProfile.objects.create(user=user)
        cls.project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)
        cls.file_uploader = FileUploader.objects.create(
            project=cls.project,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE
        )
        cls.blueprint = DonationBlueprint.objects.create(
            project=cls.project,
            name='donation blueprint',
            expected_fields='"title", "titleUrl"',
            file_uploader=None
        )
        cls.participant = Participant.objects.create(
            project=cls.project,
            start_time=timezone.now()
        )

    @staticmethod
    def generate_profiler_table(stats):
        """ Generates a pandas dataframe from cProfiler stats. """

        def filter_row(r):
            """ Keeps row only if the recorded function is part of DDM. """
            if 'code' in str(type(r['func'])):
                source_file = inspect.getsourcefile(r['func'])
                if source_file:
                    if '\\lib\\' in source_file:
                        return False
                    else:
                        return True
            return False

        df = pd.DataFrame(
            stats,
            columns=['func', 'ncalls', 'ccalls', 'tottime', 'inlinetime', 'callers']
        )
        return df[df.apply(lambda x: filter_row(x), axis=1)]

    @staticmethod
    def print_result_to_console(stats, header):
        """
        Print execution time and the 5 functions with the largest inline times.
        """
        print('------------------------------------------------------')
        print(header)
        print(f'  Execution time: {stats["tottime"].max()}')
        print(f'  Top 5 functions with longest execution time:')
        for k, r in stats.nlargest(5, 'inlinetime').iterrows():
            print(f'     Function:    {r["func"]}')
            print(f'     Inline time: {r["inlinetime"]}')
            print(f'     Ncalls:      {r["ncalls"]}; ccalls: {r["ccalls"]}')
        return


class TestDonationBlueprintModel(PerformanceTest):
    def test_blueprint_donation_save_time(self):
        """
        Test execution times to encrypt and save a donation for varying donation sizes.
        """

        # Load test data donation.
        with open('../ddm/tests/files_for_tests/test_data_1mb.json') as file:
            data_1mb = json.load(file)

        # Run execution time tests.
        times = []
        for size in [1, 5, 10, 100]:
            response = {
                'consent': True,
                'extracted_data': json.dumps(data_1mb * size),
                'status': 'some status'
            }

            with cProfile.Profile() as profiler:
                profiler.enable()
                self.blueprint.process_donation(response, self.participant)
                profiler.disable()

                # Get statistics.
                profiler_stats = PerformanceTest.generate_profiler_table(profiler.getstats())

                # Update results list.
                times.append(profiler_stats['tottime'].max())

                # Print test results to console.
                header = f'Results for donation of size {size}mb'
                PerformanceTest.print_result_to_console(profiler_stats, header)

        print(f'\nExecution times: {times}')
        return


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestDownloadAPIPerformance(PerformanceTest):

    def test_api_download_time(self):
        """ Profile blueprint processing performance. """
        with open('../ddm/tests/files_for_tests/test_data_1mb.json') as file:
            data_1mb = json.load(file)

        donation_size = 10
        donation_data = {
            'consent': True,
            'extracted_data': json.dumps(data_1mb*donation_size),
            'status': 'some status'
        }

        self.client.login(**self.user_creds)

        # Run execution time tests.
        times = []
        n_donations = 0
        n_runs = 2
        for i in range(1, n_runs + 1):
            for _ in range(10):
                DataDonation.objects.create(
                    project=self.project,
                    blueprint=self.blueprint,
                    participant=self.participant,
                    consent=donation_data['consent'],
                    status=donation_data['status'],
                    data=donation_data['extracted_data']
                )
            n_donations += 10

            with cProfile.Profile() as profiler:
                header = {'Super-Secret': None}
                profiler.enable()
                self.client.get(reverse('ddm-data-api', args=[self.project.pk]), **header)
                profiler.disable()

                # Get statistics.
                profiler_stats = PerformanceTest.generate_profiler_table(profiler.getstats())

                # Update results list.
                times.append(profiler_stats['tottime'].max())

                # Print test results to console.
                header = f'Results for run {i}/{n_runs} with {n_donations} donations à {donation_size}mb'
                PerformanceTest.print_result_to_console(profiler_stats, header)

        print(f'\nExecution times: {times}')
        return



# Generated by Django 3.2.13 on 2023-02-17 16:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0035_auto_20230217_1552'),
    ]

    operations = [
        migrations.AlterField(
            model_name='processingrule',
            name='replacement_value',
            field=models.TextField(blank=True, help_text='Only required for operation "Replace match (regex)".'),
        ),
    ]



# Generated by Django 3.2.13 on 2022-12-12 20:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0018_auto_20220929_1308'),
    ]

    operations = [
        migrations.AddField(
            model_name='donationblueprint',
            name='csv_delimiter',
            field=models.CharField(blank=True, default='', max_length=10),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='exp_file_format',
            field=models.CharField(choices=[('json', 'Json Format'), ('csv', 'Csv Format')], default='json', max_length=10, verbose_name='Expected file format'),
        ),
        migrations.AlterField(
            model_name='processingrule',
            name='comparison_value',
            field=models.TextField(blank=True),
        ),
    ]



import json

from django.views.decorators.debug import sensitive_variables
from rest_framework import serializers
from rest_framework.fields import empty

from ddm.models.core import (
    DataDonation, DonationProject, QuestionnaireResponse, Participant
)
from ddm.models.questions import QuestionBase, QuestionItem


class ProjectSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = DonationProject
        fields = ['pk', 'name', 'date_created']


class SerializerDecryptionMixin:
    """
    Allows to pass the secret for the decryption of super secret projects to
    the serializer on init.
    """
    @sensitive_variables()
    def __init__(self, instance=None, data=empty, decryptor=None, **kwargs):
        self.decryptor = decryptor
        self.secret = kwargs.pop('secret', None)
        super().__init__(instance=instance, data=data, **kwargs)

    @sensitive_variables()
    def get_data(self, obj):
        if not self.secret:
            self.secret = obj.project.secret_key
        return obj.get_decrypted_data(self.secret, obj.project.get_salt(), self.decryptor)


class DonationSerializer(SerializerDecryptionMixin, serializers.HyperlinkedModelSerializer):
    project = serializers.IntegerField(source='project.id')
    data = serializers.SerializerMethodField()
    participant = serializers.IntegerField(source='participant.id')

    class Meta:
        model = DataDonation
        fields = ['time_submitted', 'consent', 'status', 'data', 'project', 'participant']


class ResponseSerializer(SerializerDecryptionMixin, serializers.HyperlinkedModelSerializer):
    project = serializers.IntegerField(source='project.id')
    meta_data = serializers.SerializerMethodField()
    responses = serializers.SerializerMethodField()
    participant = serializers.IntegerField(source='participant.id')

    class Meta:
        model = QuestionnaireResponse
        fields = ['time_submitted', 'meta_data', 'project', 'participant', 'responses']

    @sensitive_variables()
    def get_meta_data(self, obj):
        data = super().get_data(obj)
        try:
            return json.loads(data)
        except TypeError:
            return data

    @sensitive_variables()
    def get_responses(self, obj):
        """
        Creates a dictionary that only holds 'variable_name: response' pairs.
        """
        data = super().get_data(obj)
        try:
            data = json.loads(data)
        except TypeError:
            return data

        responses = dict()
        for question_id in data:
            if isinstance(data[question_id]['response'], dict):
                item_answers = data[question_id]['response']
                for item_id in item_answers:
                    try:
                        item = QuestionItem.objects.all().get(id=item_id)
                    except QuestionItem.DoesNotExist:
                        # Item has been deleted.
                        continue
                    var_name = item.question.variable_name
                    value = item.value
                    responses[f'{var_name}-{value}'] = item_answers[item_id]
                pass
            else:
                try:
                    question = QuestionBase.objects.all().get(id=question_id)
                except QuestionBase.DoesNotExist:
                    # Question has been deleted.
                    continue
                var_name = question.variable_name
                responses[var_name] = data[question_id]['response']
        return responses


class ParticipantSerializer(serializers.HyperlinkedModelSerializer):
    project = serializers.IntegerField(source='project.id')

    class Meta:
        model = Participant
        fields = ['pk', 'project', 'external_id', 'start_time', 'end_time',
                  'completed', 'extra_data']



import re

from django.conf import settings

from ddm.models.core import DonationProject, ResearchProfile


def email_is_valid(email_string):
    """
    Check if an email address complies with the pattern specified in
    DDM_SETTINGS['EMAIL_PERMISSION_CHECK'].
    If this setting is not defined, returns True.
    """
    if hasattr(settings, 'DDM_SETTINGS'):
        if 'EMAIL_PERMISSION_CHECK' in settings.DDM_SETTINGS:
            match = re.match(settings.DDM_SETTINGS['EMAIL_PERMISSION_CHECK'],
                             email_string)
            if not match:
                return False
    return True


def user_is_permitted(user):
    """
    Check if a user has access permission.
    """
    if user.is_superuser:
        return True
    elif user.is_authenticated:
        profile = ResearchProfile.objects.filter(user=user).first()
        if profile:
            if profile.ignore_email_restriction:
                return True
        if email_is_valid(user.email):
            return True
    return False


def user_is_owner(user, project_pk):
    """
    Check if a given user is the owner of the project with the ID provided in
    project_pk.
    """
    donation_project = DonationProject.objects.get(pk=project_pk)
    if donation_project.owner is None:
        return False
    elif donation_project.owner.user == user:
        return True
    else:
        return False



from ddm.models.core import (
    DataDonation, QuestionnaireResponse, ResearchProfile, DonationProject,
    Participant, DonationBlueprint
)
from ddm.models.questions import OpenQuestion
from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from django.urls import reverse
from django.utils import timezone
from rest_framework.test import APIClient


User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestAPIs(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.base_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        }
        base_user = User.objects.create_user(**cls.base_creds)
        base_profile = ResearchProfile.objects.create(user=base_user)

        cls.no_perm_creds = {
            'username': 'non-perm', 'password': '123', 'email': 'non-perm@mail.com'
        }
        no_perm_user = User.objects.create_user(**cls.no_perm_creds)
        no_perm_profile = ResearchProfile.objects.create(user=no_perm_user)

        cls.project_base = DonationProject.objects.create(
            name='Base Project', slug='base', owner=base_profile)
        cls.project_alt = DonationProject.objects.create(
            name='Alt Project', slug='alternative', owner=no_perm_profile)

        cls.participant = Participant.objects.create(
            project=cls.project_base,
            start_time=timezone.now()
        )
        cls.participant_two = Participant.objects.create(
            project=cls.project_base,
            start_time=timezone.now()
        )
        blueprint = DonationBlueprint.objects.create(
            project=cls.project_base,
            name='donation blueprint',
            expected_fields='"a", "b"',
            file_uploader=None
        )
        DataDonation.objects.create(
            project=cls.project_base,
            blueprint=blueprint,
            participant=cls.participant,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data={'data': ['donated_data', 'donated_data']}
        )
        DataDonation.objects.create(
            project=cls.project_base,
            blueprint=blueprint,
            participant=cls.participant_two,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data={'data': ['donated_data2', 'donated_data2']}
        )
        q = OpenQuestion.objects.create(
            project=cls.project_base,
            name='open question',
            page=1,
            index=1,
            variable_name='open'
        )
        cls.q_response = QuestionnaireResponse.objects.create(
            project=cls.project_base,
            participant=cls.participant,
            time_submitted=timezone.now(),
            data=f'{{"{q.pk}": {{"response": "response_data", "question": "question text", "items": []}}}}'
        )
        cls.q_response_two = QuestionnaireResponse.objects.create(
            project=cls.project_base,
            participant=cls.participant_two,
            time_submitted=timezone.now(),
            data=f'{{"{q.pk}": {{"response": "response_data2", "question": "question text", "items": []}}}}'
        )

    def test_download_project_data_view_exists(self):
        response = self.client.get(
            reverse('ddm-data-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 401)

    def test_download_project_data_with_regular_login_owner(self):
        self.client.login(**self.base_creds)
        response = self.client.get(
            reverse('ddm-data-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 200)

    def test_download_project_data_fails_for_user_without_permission(self):
        self.client.login(**self.no_perm_creds)
        response = self.client.get(
            reverse('ddm-data-api', args=[self.project_base.pk]), follow=True)
        self.assertEqual(response.status_code, 403)

    def test_download_project_data_with_valid_api_credentials(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        response = client.get(
            reverse('ddm-data-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 200)

    def test_download_project_data_with_invalid_api_credentials(self):
        token = self.project_alt.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        response = client.get(
            reverse('ddm-data-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 401)

    def test_download_project_data_with_no_api_credentials_created(self):
        token = self.project_base.create_token()
        key = token.key
        token.delete()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + key)
        response = client.get(
            reverse('ddm-data-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 401)

    def test_delete_project_data_with_regular_login_owner(self):
        self.client.login(**self.base_creds)
        response = self.client.delete(
            reverse('ddm-delete-data', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.data)

    def test_delete_project_data_fails_for_user_without_permission(self):
        self.client.login(**self.no_perm_creds)
        response = self.client.delete(
            reverse('ddm-delete-data', args=[self.project_base.pk]), follow=True)
        self.assertEqual(response.status_code, 403)

    def test_delete_project_data_fails_with_valid_api_credentials(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        response = client.delete(
            reverse('ddm-delete-data', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 403)
        self.assertIsNotNone(response.data)

    def test_delete_project_data_fails_with_invalid_api_credentials(self):
        token = self.project_alt.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        response = client.delete(
            reverse('ddm-delete-data', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 403)

    def test_delete_project_data_with_no_api_credentials_created(self):
        token = self.project_base.create_token()
        key = token.key
        token.delete()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + key)
        response = client.delete(
            reverse('ddm-delete-data', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 403)

    def test_participant_deletion_with_regular_login(self):
        self.client.login(**self.base_creds)

        new_participant = Participant.objects.create(
            project=self.project_base, start_time=timezone.now())
        external_id = new_participant.external_id

        response = self.client.delete(
            reverse('ddm-delete-participant', args=[self.project_base.pk, external_id]))
        self.assertEqual(response.status_code, 200)
        self.assertIsNone(Participant.objects.filter(external_id=external_id).first())

        response = self.client.delete(
            reverse('ddm-delete-participant', args=[self.project_base.pk, 'some_bogus_id']))
        self.assertEqual(response.status_code, 404)

    def test_participant_deletion_fails_for_user_without_permission(self):
        self.client.login(**self.no_perm_creds)
        response = self.client.delete(
            reverse('ddm-delete-participant', args=[self.project_base.pk, self.participant.external_id]), follow=True)
        self.assertEqual(response.status_code, 403)
        self.assertIsNotNone(Participant.objects.filter(external_id=self.participant.external_id).first())

    def test_donations_api_get_with_valid_api_credentials(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        expected_response = {
            'donation blueprint': [{'data': ['donated_data', 'donated_data']},
                                   {'data': ['donated_data2', 'donated_data2']}]
        }
        response = client.get(
            reverse('donations-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, expected_response)

        expected_response = {
            'donation blueprint': [{'data': ['donated_data', 'donated_data']}]
        }
        response = client.get(
            reverse('donations-api', args=[self.project_base.pk]), {'participants': str(self.participant.external_id)})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, expected_response)

    def test_donations_api_get_fails_with_invalid_api_credentials(self):
        token = self.project_alt.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        response = client.delete(
            reverse('donations-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 401)

    def test_donations_api_fails_with_no_api_credentials_created(self):
        token = self.project_base.create_token()
        key = token.key
        token.delete()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + key)
        response = client.delete(
            reverse('donations-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 401)

    def test_responses_api_get_with_valid_api_credentials(self):
        token = self.project_base.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)

        expected_response = [{'open': 'response_data'}, {'open': 'response_data2'}]
        response = client.get(
            reverse('responses-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, expected_response)

        expected_response = [{'open': 'response_data'}]
        response = client.get(
            reverse('responses-api', args=[self.project_base.pk]), {'participants': str(self.participant.external_id)})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data, expected_response)

    def test_responses_api_get_fails_with_invalid_api_credentials(self):
        token = self.project_alt.create_token()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        response = client.delete(
            reverse('responses-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 401)

    def test_responses_api_fails_with_no_api_credentials_created(self):
        token = self.project_base.create_token()
        key = token.key
        token.delete()
        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION='Token ' + key)
        response = client.delete(
            reverse('responses-api', args=[self.project_base.pk]))
        self.assertEqual(response.status_code, 401)



# Generated by Django 3.2.13 on 2023-12-03 13:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0045_alter_processingrule_comparison_operator'),
    ]

    operations = [
        migrations.AddField(
            model_name='donationblueprint',
            name='expected_fields_regex_matching',
            field=models.BooleanField(default=False, help_text='Select if you use regex expressions in the "Excpected fields".'),
        ),
    ]



from Crypto.Cipher import PKCS1_OAEP, AES
from Crypto.Hash import HMAC
from Crypto.PublicKey import RSA
from Crypto.Random import get_random_bytes
from django.db import models
from io import BytesIO
from struct import pack

import base64
import json

from django.views.decorators.debug import sensitive_variables


class ModelWithEncryptedData(models.Model):

    class Meta:
        abstract = True

    @sensitive_variables()
    def save(self, *args, **kwargs):

        if 'encryptor' in kwargs:
            self.data = kwargs['encryptor'].encrypt(self.data)
        else:
            secret = kwargs['secret'] if 'secret' in kwargs else self.project.secret_key
            self.data = Encryption(secret, self.project.get_salt(), self.project.public_key).encrypt(self.data)
        super().save(*args, **kwargs)

    @sensitive_variables()
    def get_decrypted_data(self, secret, salt, decryptor=None):
        try:
            if decryptor:
                return decryptor.decrypt(self.data)
            else:
                return Decryption(secret, salt).decrypt(self.data)
        except ValueError as e:
            raise ValueError(f'Wrong secret, {e}')


class PRNG(object):
    def __init__(self, seed):
        self.index = 0
        self.seed = seed
        self.buffer = b""

    def __call__(self, n):
        while len(self.buffer) < n:
            self.buffer += HMAC.new(self.seed + pack("<I", self.index)).digest()
            self.index += 1
        result, self.buffer = self.buffer[:n], self.buffer[n:]
        return result


class AsyncSyncCrypto:
    @staticmethod
    def get_public_key(secret, salt):
        return Encryption.get_rsa(secret, salt).publickey().exportKey()

    @staticmethod
    def get_rsa(secret, salt):
        if secret is None:
            raise ValueError('secret is None.')
        if salt is None:
            raise ValueError('salt is None.')
        seed_128 = HMAC.new(bytes(secret, 'utf-8') + bytes(salt, 'utf-8')).digest()
        return RSA.generate(2048, randfunc=PRNG(seed_128))


class Encryption(AsyncSyncCrypto):

    def __init__(self, secret=None, salt=None, public=None):
        if public:
            self.public_key = public
        else:
            self.public_key = self.get_public_key(secret, salt)

    def encrypt(self, value):
        """
        Save encrypted data to the database, either use a supplied 'public_key'
        or generate the material dynamically.
        """
        session_key = get_random_bytes(16)
        cipher_aes = AES.new(session_key, AES.MODE_EAX)
        cipher_data, tag = cipher_aes.encrypt_and_digest(bytes(json.dumps(value), 'utf-8'))
        cipher_rsa = PKCS1_OAEP.new(RSA.importKey(self.public_key))
        encrypted_session_key = cipher_rsa.encrypt(session_key)
        payload = encrypted_session_key + cipher_aes.nonce + tag + cipher_data
        return base64.encodebytes(payload)


class Decryption(AsyncSyncCrypto):

    def __init__(self, secret, salt):
        self.rsa = self.get_rsa(secret, salt)

    def decrypt(self, value):
        """
        Fetch encrypted data from the database, try to decode the stored data,
        None if decryption fails.
        """
        value_bytes = BytesIO(base64.decodebytes(value))
        encrypted_session_key = value_bytes.read(self.rsa.size_in_bytes())
        nonce = value_bytes.read(16)
        tag = value_bytes.read(16)
        cipher_data = value_bytes.read()
        session_key = PKCS1_OAEP.new(self.rsa).decrypt(encrypted_session_key)
        cipher_aes = AES.new(session_key, AES.MODE_EAX, nonce)
        decrypted = cipher_aes.decrypt_and_verify(cipher_data, tag)
        return json.loads(decrypted.decode('utf-8'))



# Generated by Django 3.2.13 on 2023-01-22 17:28

import ddm.models.core
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0030_auto_20230120_1746'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationproject',
            name='img_header_left',
            field=models.ImageField(blank=True, null=True, upload_to=ddm.models.core.project_header_dir_path, verbose_name='Header Image Left'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='img_header_right',
            field=models.ImageField(blank=True, null=True, upload_to=ddm.models.core.project_header_dir_path, verbose_name='Header Image Right'),
        ),
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from django.urls import reverse
from django.utils import timezone
from ddm.models.core import (
    DonationProject, ResearchProfile, DonationBlueprint, FileUploader,
    DonationInstruction, Participant
)
from ddm.models.questions import (
    MatrixQuestion, SingleChoiceQuestion, OpenQuestion, MultiChoiceQuestion,
    SemanticDifferential, Transition
)


User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestAdminViewAuthentication(TestCase):
    """
    Tests that users with different access rights are redirected to
    the correct view.
    """
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        # Users
        cls.super_user_creds = {
            'username': 'superuser', 'password': '123', 'email': 'super@mail.com'
        }
        cls.super_user = User.objects.create_superuser(**cls.super_user_creds)
        ResearchProfile.objects.create(user=cls.super_user)

        cls.owner_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        }
        cls.owner = User.objects.create_user(**cls.owner_creds)
        cls.owner_profile = ResearchProfile.objects.create(user=cls.owner)

        cls.non_owner_creds = {
            'username': 'non-owner', 'password': '123', 'email': 'non-owner@mail.com'
        }
        cls.non_owner = User.objects.create_user(**cls.non_owner_creds)
        cls.non_owner_profile = ResearchProfile.objects.create(user=cls.non_owner)

        cls.not_permitted_creds = {
            'username': 'no_per', 'password': '123', 'email': 'noperm@liam.com'
        }
        cls.not_permitted_user = User.objects.create_user(**cls.not_permitted_creds)
        cls.not_permitted_profile = ResearchProfile.objects.create(user=cls.not_permitted_user)

        cls.wo_profile_creds = {
            'username': 'no_prof', 'password': '123', 'email': 'noprof@mail.com'
        }
        cls.wo_profile_user = User.objects.create_user(**cls.wo_profile_creds)

        # Projects
        cls.project_owned = DonationProject.objects.create(
            name='Owned Project', slug='owned', owner=cls.owner_profile)
        cls.project_not_owned = DonationProject.objects.create(
            name='Not Owned Project', slug='not-owned', owner=cls.non_owner_profile)

        # Blueprint
        cls.blueprint = DonationBlueprint.objects.create(
            project=cls.project_owned,
            name='donation blueprint',
            expected_fields='"a", "b"',
            file_uploader=None
        )

        # File Uploader
        cls.file_uploader = FileUploader.objects.create(
            project=cls.project_owned,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE
        )

        # Instructions
        cls.instruction = DonationInstruction.objects.create(
            text='some text',
            index=1,
            file_uploader=cls.file_uploader
        )

        # Questions
        cls.open_quest = OpenQuestion.objects.create(
            project=cls.project_owned,
            blueprint=cls.blueprint,
            name='open question',
            variable_name='open_question'
        )
        cls.sc_quest = SingleChoiceQuestion.objects.create(
            project=cls.project_owned,
            blueprint=cls.blueprint,
            name='sc question',
            variable_name='sc_question'
        )
        cls.mc_quest = MultiChoiceQuestion.objects.create(
            project=cls.project_owned,
            blueprint=cls.blueprint,
            name='mc question',
            variable_name='mc_question'
        )
        cls.matrix_quest = MatrixQuestion.objects.create(
            project=cls.project_owned,
            blueprint=cls.blueprint,
            name='matrix question',
            variable_name='matrix_question'
        )
        cls.diff_quest = SemanticDifferential.objects.create(
            project=cls.project_owned,
            blueprint=cls.blueprint,
            name='diff question',
            variable_name='diff_question'
        )
        cls.trans_quest = Transition.objects.create(
            project=cls.project_owned,
            blueprint=cls.blueprint,
            name='diff question',
            variable_name='trans_question'
        )

        # Participant
        cls.participant_base = Participant.objects.create(
            project=cls.project_owned,
            start_time=timezone.now()
        )

        # Define list of admin urls for owned project.
        cls.urls = cls.get_admin_urls(cls.project_owned.pk)

        # Define list of admin urls for non-owned project.
        cls.urls_no_perm = cls.get_admin_urls(cls.project_not_owned.pk)

    @classmethod
    def get_admin_urls(cls, project_pk):
        urls = []
        project_related_views = [
            'project-edit',
            'project-delete',
            'briefing-edit',
            'debriefing-edit',
            'data-donation-overview',
            'blueprint-create',
            'questionnaire-overview'
        ]
        for view in project_related_views:
            urls.append(reverse(view, args=[project_pk]))

        blueprint_related_views = ['blueprint-edit', 'blueprint-delete']
        for view in blueprint_related_views:
            urls.append(reverse(view, args=[project_pk, cls.blueprint.pk]))

        uploader_related_views = [
            'file-uploader-edit',
            'file-uploader-delete',
            'instruction-overview',
            'instruction-create'
        ]
        for view in uploader_related_views:
            urls.append(reverse(view, args=[project_pk, cls.file_uploader.pk]))

        instruction_related_views = ['instruction-edit', 'instruction-delete']
        for view in instruction_related_views:
            urls.append(
                reverse(view, args=[project_pk, cls.file_uploader.pk, cls.instruction.pk]))

        questions = [
            ('single_choice', cls.sc_quest.pk),
            ('multi_choice', cls.mc_quest.pk),
            ('open', cls.open_quest.pk),
            ('matrix', cls.matrix_quest.pk),
            ('semantic_diff', cls.diff_quest.pk),
            ('transition', cls.trans_quest.pk)
        ]
        for question in questions:
            urls.append(
                reverse('question-create', args=[project_pk, question[0]]))

        for view in ['question-edit', 'question-delete']:
            for question in questions:
                urls.append(
                    reverse(view, args=[project_pk, question[0], question[1]]))

        item_views = [
            reverse('question-items',
                    args=[project_pk, 'single_choice', cls.sc_quest.pk]),
            reverse('question-items',
                    args=[project_pk, 'multi_choice', cls.mc_quest.pk]),
            reverse('question-items',
                    args=[project_pk, 'matrix', cls.matrix_quest.pk]),
            reverse('question-items',
                    args=[project_pk, 'semantic_diff', cls.diff_quest.pk]),
        ]
        urls += item_views

        scale_views = [
            reverse('question-scale',
                    args=[project_pk, 'matrix', cls.matrix_quest.pk]),
            reverse('question-scale',
                    args=[project_pk, 'semantic_diff', cls.diff_quest.pk]),
        ]
        urls += scale_views
        return urls

    def test_logged_out_redirects_to_login_view(self):
        for url in self.urls:
            response = self.client.get(url, follow=True)
            self.assertRedirects(response, reverse('ddm-login'))

    def test_superuser_logged_in_returns_200(self):
        for url in self.urls:
            self.client.login(**self.super_user_creds)
            response = self.client.get(url, follow=True)
            self.assertEqual(response.status_code, 200)

    def test_owner_logged_in_returns_200(self):
        for url in self.urls:
            self.client.login(**self.owner_creds)
            response = self.client.get(url, follow=True)
            self.assertEqual(response.status_code, 200)

    def test_non_owner_logged_in_returns_404(self):
        for url in self.urls:
            if url in [reverse('project-list'), reverse('project-create')]:
                pass
            else:
                self.client.login(**self.non_owner_creds)
                response = self.client.get(url, follow=True)
                self.assertEqual(response.status_code, 404)

    def test_user_without_profile_logged_in_creates_profile_and_returns_200(self):
        for url in self.urls:
            self.assertFalse(ResearchProfile.objects.filter(user=self.wo_profile_user).exists())
            self.client.login(**self.wo_profile_creds)
            response = self.client.get(url, follow=True)
            self.assertTrue(ResearchProfile.objects.filter(user=self.wo_profile_user).exists())
            self.assertEqual(response.status_code, 200)
            ResearchProfile.objects.get(user=self.wo_profile_user).delete()

    def test_user_wo_permission_logged_in_redirects_to_no_permission_view(self):
        for url in self.urls_no_perm:
            if url in [reverse('project-list'), reverse('project-create')]:
                pass
            else:
                self.client.login(**self.not_permitted_creds)
                response = self.client.get(url, follow=True)
                self.assertRedirects(response, reverse('ddm-no-permission'))


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestProjectListView(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.owner_creds = {
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'}
        cls.owner = User.objects.create_user(**cls.owner_creds)
        cls.owner_profile = ResearchProfile.objects.create(user=cls.owner)

        # Projects
        cls.project_owned = DonationProject.objects.create(
            name='Owned Project', slug='owned', owner=cls.owner_profile)

    def test_project_list_returns_correct_queryset(self):
        self.client.login(**self.owner_creds)
        response = self.client.get(reverse('project-list'))
        object_list = response.context['object_list']
        expected_queryset = DonationProject.objects.filter(owner__user=self.owner)
        self.assertQuerysetEqual(
            object_list, list(expected_queryset), ordered=False)



# Generated by Django 3.2.9 on 2022-05-02 16:10

import ckeditor.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0003_auto_20220429_1154'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationblueprint',
            name='exp_file_format',
            field=models.CharField(choices=[('json', 'Json Format')], default='json', max_length=10, verbose_name='Expected File Format'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='intro_text',
            field=ckeditor.fields.RichTextField(blank=True, null=True, verbose_name='Welcome Page Text'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='outro_text',
            field=ckeditor.fields.RichTextField(blank=True, null=True, verbose_name='End Page Text'),
        ),
    ]



# Default django-CKEditor configuration.
ddm_ckeditor = {
    'toolbar_custom': [
        {'name': 'basicstyles',
         'items': ['Bold', 'Italic', 'Underline', 'Strike', 'Subscript', 'Superscript', '-', 'RemoveFormat']},
        {'name': 'paragraph',
         'items': ['NumberedList', 'BulletedList', '-', 'Outdent', 'Indent', '-',
                   'JustifyLeft', 'JustifyCenter', 'JustifyRight', 'JustifyBlock', '-', 'Blockquote', 'CreateDiv', '-']},
        {'name': 'links', 'items': ['Link', 'Unlink', 'Anchor']},
        {'name': 'insert',
         'items': ['Image', 'Flash', 'Table', 'HorizontalRule', 'SpecialChar']},
        {'name': 'styles', 'items': ['Styles', 'Format', 'Font', 'FontSize']},
        {'name': 'colors', 'items': ['TextColor', 'BGColor']},
        {'name': 'tools', 'items': ['Maximize', 'ShowBlocks', 'Source']}
    ],
    'toolbar': 'custom',
    'allowedContent': True,
    'removePlugins': 'exportpdf',
    'entities': False,
    'extraPlugins': 'uploadimage',
}



# Generated by Django 3.2.13 on 2023-03-08 15:19

import ckeditor.fields
import ckeditor_uploader.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0037_auto_20230227_1117'),
    ]

    operations = [
        migrations.AddField(
            model_name='donationblueprint',
            name='json_extraction_root',
            field=models.CharField(blank=True, default='', max_length=200),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='description',
            field=models.TextField(help_text='A description of which kind of data will be extracted by this Blueprint (e.g., "The title of your watched videos will be collected together with the time when you watched it."). Will be visible to participants.', null=True),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='file_uploader',
            field=models.ForeignKey(blank=True, help_text='The File Uploader through which the related file will be uploaded.', null=True, on_delete=django.db.models.deletion.SET_NULL, to='ddm.fileuploader', verbose_name='Associated File Uploader'),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='name',
            field=models.CharField(help_text='Name for this File Blueprint. Will be visible to participants, so pick an informative name (e.g., "Watch History").', max_length=250),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='regex_path',
            field=models.TextField(blank=True, help_text='The path where the file is expected to be located in the uploaded ZIP folder. You can use Regex to, e.g., add wildcard characters or to match files in different languages. Consult the documentation for some examples.', null=True, verbose_name='File path'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='briefing_consent_enabled',
            field=models.BooleanField(default=False, help_text='Enable this option to obtain explicit consent from participant, that they want to start the study.', verbose_name='Briefing Consent Mandatory'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='briefing_text',
            field=ckeditor_uploader.fields.RichTextUploadingField(blank=True, help_text='This text will be displayed to participants in the first step before beginning the data donation. Here, you should introduce your study, who is responsible for the study, provide a quick summary of what your participants will be asked to do in the next steps, etc.', null=True, verbose_name='Briefing Text'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='contact_information',
            field=ckeditor.fields.RichTextField(help_text='Please provide the contact information of the person responsible for the conduction of this study (Name, professional address, e-mail, tel, ...). The contact information will be accessible for participants during the data donation. The field is mandatory.', null=True, verbose_name='Contact information'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='data_protection_statement',
            field=ckeditor.fields.RichTextField(help_text='Please provide a data protection statement for your data donation collection. This should include the purpose for which the data is collected, how it will be stored, and who will have access to the data. The field is mandatory.', null=True, verbose_name='Data Protection Statement'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='debriefing_text',
            field=ckeditor_uploader.fields.RichTextUploadingField(blank=True, help_text='This text will be displayed to participants on the last page of the study (i.e., after the data donation and the questionnaire).', null=True, verbose_name='Debriefing Text'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='name',
            field=models.CharField(help_text='Project Name - for internal organisation only (can still be changed later).', max_length=50),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='slug',
            field=models.SlugField(help_text='Identifier that is included in the URL through which participants can access the project (e.g, https://root.url/url-identifier).', unique=True, verbose_name='URL Identifier'),
        ),
        migrations.AlterField(
            model_name='processingrule',
            name='comparison_operator',
            field=models.CharField(blank=True, choices=[('==', 'Equal (==)'), ('!=', 'Not Equal (!=)'), ('>', 'Greater than (>)'), ('<', 'Smaller than (<)'), ('>=', 'Greater than or equal (>=)'), ('<=', 'Smaller than or equal (<=)'), ('regex-delete-match', 'Delete match (regex)'), ('regex-replace-match', 'Replace match (regex)'), ('regex-delete-row', 'Delete row when match (regex)')], default=None, max_length=24, null=True, verbose_name='Extraction Operator'),
        ),
        migrations.AlterField(
            model_name='processingrule',
            name='comparison_value',
            field=models.TextField(blank=True, help_text='The value against which the data contained in the indicated field will be compared according to the selected comparison logic.'),
        ),
        migrations.AlterField(
            model_name='processingrule',
            name='execution_order',
            field=models.IntegerField(help_text='The order in which the extraction steps are executed.'),
        ),
        migrations.AlterField(
            model_name='processingrule',
            name='field',
            field=models.TextField(help_text='The field on which the rule will be applied (just as a string without quotes).If a field is mentioned in a rule, it will be kept in the data that are sent to the server.'),
        ),
        migrations.AlterField(
            model_name='processingrule',
            name='name',
            field=models.CharField(help_text='An informative name for this rule. Only used internally.', max_length=250),
        ),
        migrations.AlterField(
            model_name='questionbase',
            name='question_type',
            field=models.CharField(choices=[('generic', 'Generic Question'), ('single_choice', 'Single Choice Question'), ('multi_choice', 'Multi Choice Question'), ('matrix', 'Matrix Question'), ('semantic_diff', 'Semantic Differential'), ('open', 'Open Question'), ('transition', 'Text Block')], default='generic', max_length=20),
        ),
        migrations.AlterField(
            model_name='questionbase',
            name='text',
            field=ckeditor_uploader.fields.RichTextUploadingField(blank=True, help_text='If a question is linked to a File Blueprint, data points from the donated data associated with the linked donation blueprint can be included in the question text. This data can be included as "{{ data }}" in the question text. It is possible to subset the data object (e.g., to include the last datapoint you can use {{ data.0 }} or include advanced rendering options included in the Django templating engine. For a more comprehensive overview and examples see [LINK TO BE INCLUDED]. Additionally, information directly related to the participant can be included in the question text. This information can be referenced as "{{ participant }}".', null=True),
        ),
    ]



# Generated by Django 3.2.13 on 2023-04-29 20:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0041_auto_20230429_1258'),
    ]

    operations = [
        migrations.AddField(
            model_name='questionbase',
            name='page',
            field=models.PositiveIntegerField(default=1, verbose_name='Page'),
        ),
        migrations.AlterField(
            model_name='questionbase',
            name='index',
            field=models.PositiveIntegerField(default=1, verbose_name='Index'),
        ),
    ]



# Generated by Django 3.2.13 on 2023-02-01 14:55

import django.core.validators
import random
import string
from django.db import migrations, models


def generate_unique_external_ids(apps, schema_editor):
    participant_model = apps.get_model('ddm', 'participant')
    for p in participant_model.objects.all():
        p.external_id = ''.join(random.choices(string.ascii_letters + string.digits, k=24))
        p.save()


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0032_auto_20230130_1733'),
    ]

    operations = [
        migrations.AddField(
            model_name='participant',
            name='external_id',
            field=models.CharField(default='rSnmWUWmKXhGwGrFEqVeM9c2', max_length=24, null=True, validators=[django.core.validators.MinLengthValidator(24)]),
        ),
        migrations.RunPython(generate_unique_external_ids),
        migrations.AlterField(
            model_name='participant',
            name='external_id',
            field=models.CharField(max_length=24, unique=True,
                                   validators=[django.core.validators.MinLengthValidator(24)]),
        ),
    ]



# Generated by Django 3.2.13 on 2023-01-12 10:40

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0027_auto_20230110_1738'),
    ]

    operations = [
        migrations.AlterField(
            model_name='exceptionlogentry',
            name='date',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.CreateModel(
            name='EventLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(default=django.utils.timezone.now)),
                ('description', models.CharField(max_length=100)),
                ('message', models.TextField()),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
            ],
        ),
    ]



# Generated by Django 3.2.13 on 2022-09-22 11:24

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0012_auto_20220921_1634'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='ZippedBlueprint',
            new_name='BlueprintContainer',
        ),
        migrations.RemoveConstraint(
            model_name='donationinstruction',
            name='unique_index_per_zipblueprint',
        ),
        migrations.RenameField(
            model_name='donationinstruction',
            old_name='zip_blueprint',
            new_name='blueprint_container',
        ),
        migrations.RemoveField(
            model_name='donationblueprint',
            name='zip_blueprint',
        ),
        migrations.AddField(
            model_name='donationblueprint',
            name='blueprint_container',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='ddm.blueprintcontainer', verbose_name='Blueprint container'),
        ),
        migrations.AddConstraint(
            model_name='donationinstruction',
            constraint=models.UniqueConstraint(fields=('index', 'blueprint_container'), name='unique_index_per_container'),
        ),
    ]



# Generated by Django 3.2.13 on 2022-09-21 14:34

import ckeditor.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0011_auto_20220830_1047'),
    ]

    operations = [
        migrations.AddField(
            model_name='donationproject',
            name='contact_information',
            field=ckeditor.fields.RichTextField(null=True, verbose_name='Contact information'),
        ),
        migrations.AddField(
            model_name='donationproject',
            name='data_protection_statement',
            field=ckeditor.fields.RichTextField(null=True, verbose_name='Data Protection Statement'),
        ),
    ]



# Generated by Django 3.2.13 on 2022-06-21 09:36

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('ddm', '0006_auto_20220617_1433'),
    ]

    operations = [
        migrations.CreateModel(
            name='ResearchProfile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('active', models.BooleanField(default=True)),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date registered')),
                ('ignore_email_restriction', models.BooleanField(default=False)),
                ('api_key', models.CharField(blank=True, max_length=16, null=True, unique=True)),
                ('api_secret', models.CharField(blank=True, max_length=64, null=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='ddm_research_profile', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddField(
            model_name='donationproject',
            name='collaborators',
            field=models.ManyToManyField(related_name='project_collaborators', to='ddm.ResearchProfile'),
        ),
        migrations.AddField(
            model_name='donationproject',
            name='owner',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='project_owner', to='ddm.researchprofile'),
        ),
    ]



# Generated by Django 3.2.13 on 2023-12-03 11:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0043_auto_20230525_1621'),
    ]

    operations = [
        migrations.AddField(
            model_name='processingrule',
            name='regex_field',
            field=models.BooleanField(default=False, help_text='Select if you use a regex expression in the "Field" setting to match a variable.'),
        ),
        migrations.AlterField(
            model_name='openquestion',
            name='display',
            field=models.CharField(choices=[('small', 'Small'), ('large', 'Large')], default='large', help_text='"Small" displays a one-line textfield, "Large" a multiline textfield as input.', max_length=20),
        ),
    ]



# Generated by Django 3.2.13 on 2023-02-10 10:48

import ckeditor_uploader.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0033_participant_external_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationinstruction',
            name='text',
            field=ckeditor_uploader.fields.RichTextUploadingField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='briefing_text',
            field=ckeditor_uploader.fields.RichTextUploadingField(blank=True, null=True, verbose_name='Welcome Page Text'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='debriefing_text',
            field=ckeditor_uploader.fields.RichTextUploadingField(blank=True, null=True, verbose_name='End Page Text'),
        ),
        migrations.AlterField(
            model_name='questionbase',
            name='text',
            field=ckeditor_uploader.fields.RichTextUploadingField(blank=True, null=True),
        ),
    ]



# Generated by Django 3.2.13 on 2023-03-09 08:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0038_auto_20230308_1619'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationblueprint',
            name='json_extraction_root',
            field=models.CharField(blank=True, default='', help_text="Indicates on which level of the files' data structure information should be extractet. If you want to extract information contained on the first level (e.g., {'field to be extracted': value}, you can leave this field empty. If you want to extract data located on a higher level, then you would provide the path to the parent field of the data you want to extract (e.g., if your json file is structured like this {'friends': {'real_friends': [{'name to extract': name, 'date to extract': date}], 'fake friends': [{'name': name, 'date': date }]}} and you want to extract the names and dates of real_friends, you would set the extraction root to 'friends.real_friends'.", max_length=200, verbose_name='Extraction Root'),
        ),
    ]



# Generated by Django 3.2.13 on 2022-08-30 08:47

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0010_auto_20220629_1651'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationblueprint',
            name='exp_file_format',
            field=models.CharField(choices=[('json', 'Json Format')], default='json', max_length=10, verbose_name='Expected file format'),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='expected_fields',
            field=models.TextField(help_text='Put the field names in double quotes (") and separate them with commas ("Field A", "Field B").', validators=[django.core.validators.RegexValidator('^((["][^"]+["]))(\\s*,\\s*((["][^"]+["])))*[,\\s]*$', message='Field must contain one or multiple comma separated strings. Strings must be enclosed in double quotes ("string").')]),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='extracted_fields',
            field=models.TextField(blank=True, help_text='Put the field names in double quotes (") and separate them with commas ("Field A", "Field B").', null=True, validators=[django.core.validators.RegexValidator('^((["][^"]+["]))(\\s*,\\s*((["][^"]+["])))*[,\\s]*$', message='Field must contain one or multiple comma separated strings. Strings must be enclosed in double quotes ("string").')]),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='zip_blueprint',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='ddm.zippedblueprint', verbose_name='Zip container'),
        ),
        migrations.AlterField(
            model_name='donationinstruction',
            name='index',
            field=models.PositiveIntegerField(default=1, validators=[django.core.validators.MinValueValidator(1)]),
        ),
        migrations.CreateModel(
            name='ExceptionLogEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField()),
                ('exception_type', models.IntegerField()),
                ('message', models.TextField()),
                ('participant', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='ddm.participant')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject')),
            ],
        ),
    ]



# Generated by Django 3.2.13 on 2023-01-12 12:30

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0028_auto_20230112_1140'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='EventLog',
            new_name='EventLogEntry',
        ),
    ]



from django.apps import AppConfig


class DdmConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'ddm'
    verbose_name = 'Data Donation Module'

    def ready(self):
        from ddm import signals

        # Add default settings for DDM if they are not defined in main application's settings.py.
        from django.conf import settings
        from ddm import default_settings as defaults

        if not hasattr(settings, 'CKEDITOR_CONFIGS'):
            setattr(settings, 'CKEDITOR_CONFIGS', {})
        if 'ddm_ckeditor' not in settings.CKEDITOR_CONFIGS:
            settings.CKEDITOR_CONFIGS['ddm_ckeditor'] = defaults.ddm_ckeditor



from django.urls import path, include
from django.views.generic import RedirectView

from ddm.views import admin, participation_flow
from ddm.views.apis import (
    ExceptionAPI, ProjectDataAPI, DeleteParticipantAPI, DeleteProjectData,
    DonationsAPI, ResponsesAPI
)


participation_flow_patterns = [
    path(r'', participation_flow.participation_redirect_view, name='participation-redirect'),
    path(r'briefing/', participation_flow.BriefingView.as_view(), name='briefing'),
    path(r'data-donation/', participation_flow.DataDonationView.as_view(), name='data-donation'),
    path(r'questionnaire/', participation_flow.QuestionnaireView.as_view(), name='questionnaire'),
    path(r'debriefing/', participation_flow.DebriefingView.as_view(), name='debriefing'),
    path(r'continue/', participation_flow.ContinuationView.as_view(), name='continuation')
]

question_patterns = [
    path(r'', admin.QuestionnaireOverview.as_view(), name='questionnaire-overview'),
    path(r'<slug:question_type>/create/', admin.QuestionCreate.as_view(), name='question-create'),
    path(r'<slug:question_type>/<int:pk>/edit/', admin.QuestionEdit.as_view(), name='question-edit'),
    path(r'<slug:question_type>/<int:pk>/delete/', admin.QuestionDelete.as_view(), name='question-delete'),
    path(r'<slug:question_type>/<int:pk>/items/', admin.ItemEdit.as_view(), name='question-items'),
    path(r'<slug:question_type>/<int:pk>/scale/', admin.ScaleEdit.as_view(), name='question-scale'),
]

instruction_patterns = [
    path(r'', admin.InstructionOverview.as_view(), name='instruction-overview'),
    path(r'create/', admin.InstructionCreate.as_view(), name='instruction-create'),
    path(r'<int:pk>/edit/', admin.InstructionEdit.as_view(), name='instruction-edit'),
    path(r'<int:pk>/delete/', admin.InstructionDelete.as_view(), name='instruction-delete'),
]

data_donation_patterns = [
    path(r'', admin.data_donations.DataDonationOverview.as_view(), name='data-donation-overview'),
    path(r'blueprint/create/', admin.BlueprintCreate.as_view(), name='blueprint-create'),
    path(r'blueprint/<int:pk>/edit/', admin.BlueprintEdit.as_view(), name='blueprint-edit'),
    path(r'blueprint/<int:pk>/delete/', admin.BlueprintDelete.as_view(), name='blueprint-delete'),
    path(r'file-uploader/create/', admin.FileUploaderCreate.as_view(), name='file-uploader-create'),
    path(r'file-uploader/<int:pk>/edit/', admin.FileUploaderEdit.as_view(), name='file-uploader-edit'),
    path(r'file-uploader/<int:pk>/delete/', admin.FileUploaderDelete.as_view(), name='file-uploader-delete'),
    path(r'file-uploader/<int:file_uploader_pk>/instructions/', include(instruction_patterns)),
]

admin_patterns = [
    path(r'', admin.ProjectList.as_view(), name='project-list'),
    path(r'create/', admin.ProjectCreate.as_view(), name='project-create'),
    path(r'<int:pk>/', admin.ProjectDetail.as_view(), name='project-detail'),
    path(r'<int:pk>/edit/', admin.ProjectEdit.as_view(), name='project-edit'),
    path(r'<int:pk>/delete/', admin.ProjectDelete.as_view(), name='project-delete'),
    path(r'<int:pk>/briefing/', admin.BriefingEdit.as_view(), name='briefing-edit'),
    path(r'<int:pk>/debriefing/', admin.DebriefingEdit.as_view(), name='debriefing-edit'),
    path(r'<int:pk>/token/', admin.ProjectAPITokenView.as_view(), name='project-token'),
    path(r'<int:project_pk>/questionnaire/', include(question_patterns)),
    path(r'<int:project_pk>/data-donation/', include(data_donation_patterns)),
    path(r'<int:project_pk>/logs/', admin.ProjectLogsView.as_view(), name='project-logs'),
]

authentication_patterns = [
    path(r'no-permission/', admin.DdmNoPermissionView.as_view(), name='ddm-no-permission'),
]

api_patterns = [
    path('project/<int:pk>/data', ProjectDataAPI.as_view(), name='ddm-data-api'),
    path('project/<int:pk>/data/delete', DeleteProjectData.as_view(), name='ddm-delete-data'),
    path('project/<int:pk>/donations', DonationsAPI.as_view(), name='donations-api'),
    path('project/<int:pk>/responses', ResponsesAPI.as_view(), name='responses-api'),
    path('project/<int:pk>/participant/<slug:participant_id>/delete', DeleteParticipantAPI.as_view(), name='ddm-delete-participant')
]

urlpatterns = [
    path(r'', RedirectView.as_view(pattern_name='project-list'), name='ddm-landing-page'),
    path(r'studies/<slug:slug>/', include(participation_flow_patterns)),
    path(r'projects/', include(admin_patterns)),
    path(r'auth/', include(authentication_patterns)),
    path(r'<int:pk>/exceptions/', ExceptionAPI.as_view(), name='ddm-exceptions-api'),
    path(r'api/', include(api_patterns))
]



import datetime
import json
import os
import random
import string

from ckeditor.fields import RichTextField
from ckeditor_uploader.fields import RichTextUploadingField

from django.conf import settings
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.core.validators import RegexValidator, MinValueValidator, MinLengthValidator
from django.db import models
from django.db.models import Avg, F, ImageField
from django.template import Context, Template
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.views.decorators.debug import sensitive_variables

from ddm.models.auth import ProjectAccessToken
from ddm.models.encryption import Encryption, ModelWithEncryptedData
from ddm.models.logs import ExceptionLogEntry, ExceptionRaisers, EventLogEntry


class ResearchProfile(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='ddm_research_profile',
    )
    active = models.BooleanField(default=True)
    created = models.DateTimeField('date registered', default=timezone.now)
    ignore_email_restriction = models.BooleanField(default=False)

    def __str__(self):
        return self.user.email


def project_header_dir_path(instance, filename):
    return f'project_{instance.pk}/headers/{filename}'


class DonationProject(models.Model):
    # Basic information for internal organization.
    name = models.CharField(
        max_length=50,
        help_text='Project Name - for internal organisation only (can still be changed later).'
    )

    # Note: Value of "date_created" attribute must not be changed after
    # instance creation as it is used for en-/decryption.
    date_created = models.DateTimeField(default=timezone.now)
    owner = models.ForeignKey(
        'ResearchProfile',
        related_name='project_owner',
        on_delete=models.SET_NULL,
        null=True
    )

    # Public information.
    contact_information = RichTextField(
        null=True,
        blank=False,
        verbose_name='Contact information',
        help_text=(
            'Please provide the contact information of the person responsible '
            'for the conduction of this study (Name, professional address, e-mail, tel, ...). The contact information will be '
            'accessible for participants during the data donation. The field is mandatory.'
        ),
        config_name='ddm_ckeditor'
    )
    data_protection_statement = RichTextField(
        null=True,
        blank=False,
        verbose_name='Data Protection Statement',
        help_text=(
            'Please provide a data protection statement for your data donation '
            'collection. This should include the purpose for which the data is '
            'collected, how it will be stored, and who will have access to the data. '
            'The field is mandatory.'
        ),
        config_name='ddm_ckeditor'
    )

    # Information affecting participation flow.
    slug = models.SlugField(
        unique=True,
        verbose_name='URL Identifier',
        help_text='Identifier that is included in the URL through which '
                  'participants can access the project '
                  '(e.g, https://root.url/url-identifier).'
    )
    briefing_text = RichTextUploadingField(
        null=True, blank=True,
        verbose_name='Briefing Text',
        config_name='ddm_ckeditor',
        help_text=(
            'This text will be displayed to participants in the first step before '
            'beginning the data donation. Here, you should introduce your study, '
            'who is responsible for the study, provide a quick summary of what '
            'your participants will be asked to do in the next steps, etc.'
        )
    )
    briefing_consent_enabled = models.BooleanField(
        default=False,
        verbose_name='Briefing Consent Mandatory',
        help_text='Enable this option to obtain explicit consent from participant, '
                  'that they want to start the study.'
    )
    briefing_consent_label_yes = models.CharField(max_length=255, blank=True)
    briefing_consent_label_no = models.CharField(max_length=255, blank=True)
    debriefing_text = RichTextUploadingField(
        null=True, blank=True,
        verbose_name='Debriefing Text',
        config_name='ddm_ckeditor',
        help_text='This text will be displayed to participants on the last page '
                  'of the study (i.e., after the data donation and the questionnaire).'
    )

    # Appearance settings.
    img_header_left = ImageField(
        upload_to=project_header_dir_path,
        null=True,
        blank=True,
        verbose_name='Header Image Left'
    )
    img_header_right = ImageField(
        upload_to=project_header_dir_path,
        null=True,
        blank=True,
        verbose_name='Header Image Right'
    )

    # Access settings.
    public_key = models.BinaryField()
    super_secret = models.BooleanField(default=False)

    # Redirect settings.
    redirect_enabled = models.BooleanField(default=False, verbose_name='Redirect enabled')
    redirect_target = models.CharField(
        max_length=2000,
        blank=True,
        verbose_name='Redirect target',
        help_text=mark_safe(
            'Always include <i>http://</i> or <i>https://</i> in the redirect target. '
            'If URL parameter extraction is enabled for this project, you can '
            'include the extracted URL parameters in the redirect target as follows: '
            '"https://redirect.me/?redirectpara=<b>{{participant.data.url_param.URLParameter}}</b>".')
    )

    # URL parameter extraction settings.
    url_parameter_enabled = models.BooleanField(default=False, verbose_name='URL parameter extraction enabled')
    expected_url_parameters = models.CharField(
        max_length=500,
        blank=True,
        verbose_name='Expected URL parameter',
        help_text='Separate multiple parameters with a semikolon (";"). '
                  'Semikolons are not allowed as part of the expected url parameters.'
    )

    @sensitive_variables()
    def __init__(self, *args, **kwargs):
        self.secret_key = settings.SECRET_KEY
        if 'secret_key' in kwargs:
            self.secret_key = kwargs['secret_key']
            kwargs.pop('secret_key')
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('project-detail', args=[str(self.id)])

    def get_salt(self):
        return str(self.date_created)

    def clean_file_on_reupload(self):
        """
        Deletes previous file on model FileFields or ImageFields if a file
        is re-uploaded.
        """
        i = DonationProject.objects.get(pk=self.pk)
        for field in i._meta.fields:
            if field.get_internal_type() in ['FileField', 'ImageField']:
                if getattr(i, field.name) != getattr(self, field.name):
                    file_attr = getattr(i, field.name)
                    try:
                        if os.path.isfile(file_attr.path):
                            os.remove(file_attr.path)
                    except ValueError:
                        pass
        return

    @sensitive_variables()
    def save(self, *args, **kwargs):
        if not self.pk:
            if self.owner is None:
                raise ValidationError(
                    'DonationProject.owner cannot be null on create.')
        else:
            self.clean_file_on_reupload()

        if not self.public_key:
            self.public_key = Encryption.get_public_key(self.secret, str(self.date_created))
        super().save(*args, **kwargs)

    @property
    def secret(self):
        return self.secret_key

    @secret.setter
    def secret(self, value):
        self.secret_key = value

    def delete(self, using=None, keep_parents=False):
        # Manually delete related questions to circumvent foreign key integrity
        # problem caused by QuestionBase models inheriting from django-polymorphic.
        # (see https://github.com/django-polymorphic/django-polymorphic/issues/34#issuecomment-1027866872)
        for question in self.questionbase_set.all():
            question.delete()
        return super().delete(using, keep_parents)

    def get_statistics(self):
        participants = Participant.objects.filter(project=self)
        statistics = {
            'n_started': participants.count(),
            'n_completed': participants.filter(completed=True).count(),
            'completion_rate': participants.filter(completed=True).count() / participants.count() if participants.count() > 0 else 0,
            'n_donations': DataDonation.objects.filter(project=self, status='success').count(),
            'n_errors': ExceptionLogEntry.objects.filter(project=self).count(),
            'average_time': str(participants.filter(completed=True).aggregate(v=Avg(F('end_time')-F('start_time')))['v']).split(".")[0]
        }
        return statistics

    def get_questionnaire_config(self, participant, view):
        """
        Returns a dictionary containing all information to render the
        questionnaire for a given participant.
        """
        q_config = []
        questions = self.questionbase_set.all().order_by('page', 'index')
        for question in questions:
            if question.is_general():
                q_config.append(question.get_config(participant, view))
            else:
                try:
                    donation = DataDonation.objects.get(
                        blueprint=question.blueprint,
                        participant=participant
                    )
                except ObjectDoesNotExist:
                    msg = ('Questionnaire Rendering Exception: No donation '
                           f'found for participant {participant.pk} and '
                           f'blueprint {question.blueprint.pk}.')
                    ExceptionLogEntry.objects.create(
                        project=self,
                        raised_by=ExceptionRaisers.SERVER,
                        message=msg
                    )
                    continue

                if donation.consent and donation.status == 'success':
                    q_config.append(question.get_config(participant, view))
        return q_config

    def get_expected_url_parameters(self):
        return self.expected_url_parameters.split(';')

    def get_token(self):
        return ProjectAccessToken.objects.filter(project=self).first()

    def create_token(self, expiration_days=None):
        token = self.get_token()
        if token:
            token.delete()
        if expiration_days:
            expiration_date = timezone.now() + datetime.timedelta(days=expiration_days)
        else:
            expiration_date = None
        EventLogEntry.objects.create(
            project=self, description='New Access Token Created')
        return ProjectAccessToken.objects.create(
            project=self, expiration_date=expiration_date)


def get_extra_data_default():
    """ Return default value for Participant.extra_data. """
    return dict(url_param=dict())


def create_asciidigits_id():
    return ''.join(random.choices(string.ascii_letters + string.digits, k=24))


class Participant(models.Model):
    project = models.ForeignKey('DonationProject', on_delete=models.CASCADE)

    external_id = models.CharField(
        unique=True, null=False,
        max_length=24,
        validators=[MinLengthValidator(24)]
    )

    # Participation statistics.
    start_time = models.DateTimeField()
    end_time = models.DateTimeField(null=True)
    completed = models.BooleanField(default=False)
    current_step = models.IntegerField(blank=True, null=True)

    extra_data = models.JSONField(default=get_extra_data_default)

    def get_context_data(self):
        """
        Returns data that can be accessed when participant is passed to a
        template as a context variable.
        """
        context_data = {
            'public_id': self.external_id,
            'data': self.extra_data,
            'donation_info': self.get_donation_info()
        }
        return context_data

    def get_donation_info(self):
        donations = DataDonation.objects.filter(participant=self)
        donation_info = {
            'n_success': donations.filter(status='success').count(),
            'n_pending': donations.filter(status='pending').count(),
            'n_failed': donations.filter(status='failed').count(),
            'n_consent': donations.filter(status='success', consent=True).count(),
            'n_no_consent': donations.filter(status='success', consent=False).count(),
            'n_no_data_extracted': donations.filter(status='nothing extracted').count()
        }
        return donation_info

    def save(self, *args, **kwargs):
        if self.pk is None:
            new_external_id = create_asciidigits_id()
            while len(Participant.objects.filter(external_id=new_external_id)) != 0:
                new_external_id = create_asciidigits_id()
            self.external_id = new_external_id
        super().save(*args, **kwargs)


class QuestionnaireResponse(ModelWithEncryptedData):
    # Will only ever be deleted, when the project is deleted.
    project = models.ForeignKey('DonationProject', on_delete=models.CASCADE)
    participant = models.ForeignKey('Participant', on_delete=models.CASCADE)
    time_submitted = models.DateTimeField(default=timezone.now)
    data = models.BinaryField()


COMMA_SEPARATED_STRINGS_VALIDATOR = RegexValidator(
    r'^((["][^"]+["]))(\s*,\s*((["][^"]+["])))*[,\s]*$',
    message=(
        'Field must contain one or multiple comma separated strings. '
        'Strings must be enclosed in double quotes ("string").'
    )
)


class FileUploader(models.Model):
    name = models.CharField(max_length=250)
    project = models.ForeignKey('DonationProject', on_delete=models.CASCADE)
    index = models.PositiveIntegerField()

    class UploadTypes(models.TextChoices):
        ZIP_FILE = 'zip file'
        SINGLE_FILE = 'single file'

    upload_type = models.CharField(
        max_length=20,
        choices=UploadTypes.choices,
        default=UploadTypes.SINGLE_FILE,
        verbose_name='Upload type',
    )

    combined_consent = models.BooleanField(
        default=False,
        verbose_name='All-in-one consent',
        help_text='If enabled, participants will be asked to consent to submit '
                  'all uploaded data at once. Otherwise, participant will be asked to '
                  'consent to the submission of the data separately for each blueprint.'
    )

    def __str__(self):
        return self.name

    def delete(self, *args, **kwargs):
        """ This model has a post_delete signal processor (see signals.py). """
        super().delete(*args, **kwargs)

    def get_configs(self, participant_data=None):
        blueprints = self.donationblueprint_set.all()
        instructions = self.donationinstruction_set.all()
        configs = {
            'upload_type': self.upload_type,
            'name': self.name,
            'combined_consent': self.combined_consent,
            'blueprints': [bp.get_config() for bp in blueprints],
            'instructions': [{
                'index': i.index,
                'text': Template(i.text).render(Context({'participant': participant_data}))
            } for i in instructions]
        }
        return configs

    def save(self, *args, **kwargs):
        if self.index is None:
            uploader_indices = self.project.fileuploader_set.all().values_list('index', flat=True)
            self.index = 1 if not uploader_indices else max(uploader_indices) + 1
        super().save(*args, **kwargs)


class DonationBlueprint(models.Model):
    project = models.ForeignKey(
        'DonationProject',
        on_delete=models.CASCADE
    )
    name = models.CharField(
        max_length=250,
        help_text=(
            'Name for this File Blueprint. Will be visible to participants, '
            'so pick an informative name (e.g., "Watch History").'
        )
    )
    description = models.TextField(
        null=True,
        help_text=(
            'A description of which kind of data will be extracted by this '
            'Blueprint (e.g., "The title of your watched videos will be '
            'collected together with the time when you watched it."). '
            'Will be visible to participants.'
        )
    )

    class FileFormats(models.TextChoices):
        JSON_FORMAT = 'json', 'JSON file'
        CSV_FORMAT = 'csv', 'CSV file'
        # HTML_FORMAT = 'html',
        # XLSX_FORMAT = 'xlsx',

    exp_file_format = models.CharField(
        max_length=10,
        choices=FileFormats.choices,
        default=FileFormats.JSON_FORMAT,
        verbose_name='Expected file format',
    )

    json_extraction_root = models.CharField(
        max_length=200,
        default='',
        blank=True,
        verbose_name='Extraction Root',
        help_text=(
            'Indicates on which level of the files\' data structure information should be extractet. '
            'If you want to extract information contained on the first level (e.g., {\'field to be extracted\': value}, '
            'you can leave this field empty. If you want to extract data located on a higher level, then you would '
            'provide the path to the parent field of the data you want to extract (e.g., if your json file is structured like this '
            '{\'friends\': {\'real_friends\': [{\'name to extract\': name, \'date to extract\': date}], \'fake friends\': [{\'name\': name, \'date\': date }]}} '
            'and you want to extract the names and dates of real_friends, you would set the extraction root to \'friends.real_friends\'.'
        )
    )

    csv_delimiter = models.CharField(
        max_length=10,
        default="",
        blank=True
    )

    expected_fields = models.TextField(
        null=False,
        blank=False,
        validators=[COMMA_SEPARATED_STRINGS_VALIDATOR],
        help_text=(
            'Put the field names in double quotes (") and separate them '
            'with commas ("Field A", "Field B").'
        )
    )

    expected_fields_regex_matching = models.BooleanField(
        default=False,
        null=False,
        help_text='Select if you use regex expressions in the "Excpected fields".'
    )

    file_uploader = models.ForeignKey(
        'FileUploader',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name='Associated File Uploader',
        help_text=(
            'The File Uploader through which the related file will be uploaded.'
        )
    )
    regex_path = models.TextField(
        null=True,
        blank=True,
        verbose_name='File path',
        help_text=(
            'The path where the file is expected to be located in the uploaded '
            'ZIP folder. You can use Regex to, e.g., add wildcard characters or '
            'to match files in different languages. Consult the documentation '
            'for some examples.'
        )
    )

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('blueprint-edit', args=[str(self.project_id), str(self.id)])

    def get_slug(self):
        return 'blueprint'

    def get_config(self):
        config = {
            'id': self.pk,
            'name': self.name,
            'description': self.description,
            'format': self.exp_file_format,
            'json_extraction_root': self.json_extraction_root,
            'expected_fields': json.loads("[" + str(self.expected_fields) + "]"),
            'exp_fields_regex_matching': self.expected_fields_regex_matching,
            'fields_to_extract': self.get_fields_to_extract(),
            'regex_path': self.regex_path,
            'filter_rules': self.get_filter_rules(),
            'csv_delimiter': self.csv_delimiter
        }
        return config

    def get_instructions(self):
        return [{'index': i.index, 'text': i.text} for i in self.donationinstruction_set.all()]

    def get_associated_questions(self):
        return self.questionbase_set.all()

    def get_filter_rules(self):
        return [r.get_rule_config() for r in self.processingrule_set.all().order_by('execution_order')]

    def get_fields_to_extract(self):
        fields = set()
        for r in self.processingrule_set.all():
            if r.comparison_operator is None:
                fields.add(r.field)
        return list(fields)

    def process_donation(self, data, participant):
        if self.validate_donation(data):
            self.create_donation(data, participant)
        else:
            msg = ('Data Donation Processing Exception: Donation validation '
                   f'failed for blueprint {self.pk}')
            ExceptionLogEntry.objects.create(
                project=self.project,
                blueprint=self,
                raised_by=ExceptionRaisers.SERVER,
                message=msg
            )
        return

    def validate_donation(self, data):
        # Check if all expected fields are in response.
        response_fields = ['consent', 'extracted_data', 'status']
        if not all(k in data for k in response_fields):
            msg = ('Data Donation Processing Exception: Donation data for '
                   f'Donation Blueprint {self.pk} does not contain the '
                   f'expected information. Expected fields: {response_fields}; '
                   f'Present fields: {data.keys()}.')
            ExceptionLogEntry.objects.create(
                project=self.oproject,
                blueprint=self,
                raised_by=ExceptionRaisers.SERVER,
                message=msg
            )
            return False

        return True

    def create_donation(self, data, participant):
        DataDonation.objects.create(
            project=self.project,
            blueprint=self,
            participant=participant,
            consent=data['consent'],
            status=data['status'],
            data=data['extracted_data']
        )
        return


class ProcessingRule(models.Model):
    """
    A processing rule that defines how the data uploaded to VUE will be processed
    before being sent to the server.
    Generates a json configuration that is passed to the VUE frontend component
    'UploaderApp'.
    """
    blueprint = models.ForeignKey(
        'DonationBlueprint',
        null=False,
        blank=False,
        on_delete=models.CASCADE
    )

    name = models.CharField(
        max_length=250,
        help_text='An informative name for this rule. Only used internally.'
    )

    field = models.TextField(
        null=False,
        blank=False,
        help_text=(
            'The field on which the rule will be applied (just as a string without quotes).'
            'If a field is mentioned in a rule, it will be kept in the data that are sent to the server.'
        )
    )
    regex_field = models.BooleanField(
        default=False,
        null=False,
        help_text='Select if you use a regex expression in the "Field" setting to match a variable.'
    )
    execution_order = models.IntegerField(
        help_text='The order in which the extraction steps are executed.'
    )

    class ComparisonOperators(models.TextChoices):
        EMPTY = '', 'Keep Field'
        EQUAL = '==', 'Equal (==)'
        NOT_EQUAL = '!=', 'Not Equal (!=)'
        GREATER = '>', 'Greater than (>)'
        SMALLER = '<', 'Smaller than (<)'
        GREATER_OR_EQUAL = '>=', 'Greater than or equal (>=)'
        SMALLER_OR_EQUAL = '<=', 'Smaller than or equal (<=)'
        REGEX_DELETE_MATCH = 'regex-delete-match', 'Delete match (regex)'
        REGEX_REPLACE_MATCH = 'regex-replace-match', 'Replace match (regex)'
        REGEX_DELETE_ROW = 'regex-delete-row', 'Delete row when match (regex)'

    comparison_operator = models.CharField(
        max_length=24,
        blank=True,
        null=True,
        choices=ComparisonOperators.choices,
        default=None,
        verbose_name='Extraction Operator'
    )
    comparison_value = models.TextField(
        blank=True,
        help_text='The value against which the data contained in the indicated field will '
                  'be compared according to the selected comparison logic.'
    )
    replacement_value = models.TextField(
        blank=True,
        help_text='Only required for operation "Replace match (regex)".'
    )

    def get_rule_config(self):
        """
        Return a configuration dict for the processing rule:
        {
            'field': 'field_name',
            'comparison_operator': '==' | '!=' | '>' | '<' | '>=' | '<=' |
                                   'regex-delete-match' | ' regex-replace-match'
                                   'regex-delete-row' | None,
            'comparison_value': '123' | Regex-String | None,
            'replacement_value': String
        }
        """
        return {
            'field': self.field,
            'regex_field': self.regex_field,
            'comparison_operator': self.comparison_operator,
            'comparison_value': self.comparison_value,
            'replacement_value': self.replacement_value
        }


class DataDonation(ModelWithEncryptedData):
    project = models.ForeignKey(
        'DonationProject',
        on_delete=models.CASCADE
    )
    blueprint = models.ForeignKey(
        'DonationBlueprint',
        null=True,
        on_delete=models.SET_NULL
    )
    participant = models.ForeignKey(
        'Participant',
        on_delete=models.CASCADE
    )
    time_submitted = models.DateTimeField(default=timezone.now)
    consent = models.BooleanField(default=False)
    status = models.JSONField()
    data = models.BinaryField()


class DonationInstruction(models.Model):
    text = RichTextUploadingField(null=True, blank=True, config_name='ddm_ckeditor')
    index = models.PositiveIntegerField(default=1, validators=[MinValueValidator(1)])

    file_uploader = models.ForeignKey(
        'FileUploader',
        null=True,
        blank=False,
        on_delete=models.CASCADE,
        verbose_name='Associated File Uploader',
    )

    class Meta:
        ordering = ['index']
        constraints = [
            models.UniqueConstraint(
                fields=['index', 'file_uploader'],
                name='unique_index_per_file_uploader'
            ),
        ]

    def clean(self):
        # Ensure that index of instruction page is not greater than set of existing instructions + 1.
        n_instructions = self.file_uploader.donationinstruction_set.all().count()
        if self.pk:
            if self.index > n_instructions:
                raise ValidationError(
                    f'Index must be in range 1 to {n_instructions}.'
                )
        else:
            if self.index > (n_instructions + 1):
                raise ValidationError(
                    f'Index must be in range 1 to {n_instructions + 1}.'
                )
        super().clean()

    def save(self, *args, **kwargs):
        if kwargs.pop('ignore_index_check', False):
            return super().save()

        initial_index = DonationInstruction.objects.get(pk=self.pk).index if self.pk else None
        index_taken = self.file_uploader.donationinstruction_set.filter(index=self.index).exclude(pk=self.pk).exists()
        if index_taken and (self.index != initial_index):

            # Account for unique constraint by doing a "proxy"-save to free index.
            target_index = self.index
            self.index = self.file_uploader.donationinstruction_set.count() + 5
            super().save()

            # Change indices of involved objects:
            queryset = self.file_uploader.donationinstruction_set.exclude(pk=self.pk)
            if initial_index is None:
                queryset = queryset.filter(index__gte=target_index).order_by('-index')
                for q in queryset:
                    q.index += 1
                for q in queryset:
                    q.save(ignore_index_check=True)
            elif target_index < initial_index:
                queryset = queryset.filter(index__gte=target_index, index__lt=initial_index).order_by('-index')
                for q in queryset:
                    q.index += 1
                for q in queryset:
                    q.save(ignore_index_check=True)
            elif target_index > initial_index:
                queryset = queryset.filter(index__gt=initial_index, index__lte=target_index).order_by('index')
                for q in queryset:
                    q.index -= 1
                for q in queryset:
                    q.save(ignore_index_check=True)

            # Revert "proxy"-save.
            self.index = target_index
            return super().save()

        return super().save()

    def delete(self, *args, **kwargs):
        """ This model has a post_delete signal processor (see signals.py). """
        super().delete(*args, **kwargs)



# Generated by Django 3.2.13 on 2023-04-29 10:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0040_participant_current_step'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='questionitem',
            constraint=models.UniqueConstraint(fields=('index', 'question'), name='unique_item_index_per_question'),
        ),
        migrations.AddConstraint(
            model_name='questionitem',
            constraint=models.UniqueConstraint(fields=('value', 'question'), name='unique_item_value_per_question'),
        ),
    ]



# Generated by Django 3.2.13 on 2024-07-09 11:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0047_fileuploader_combined_consent'),
    ]

    operations = [
        migrations.AlterField(
            model_name='eventlogentry',
            name='description',
            field=models.TextField(),
        ),
    ]



# Generated by Django 3.2.13 on 2022-09-28 12:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0016_auto_20220928_1344'),
    ]

    operations = [
        migrations.AlterField(
            model_name='processingrule',
            name='comparison_operator',
            field=models.CharField(blank=True, choices=[('==', 'Equal (==)'), ('!=', 'Not Equal (!=)'), ('>', 'Greater than (>)'), ('<', 'Smaller than (<)'), ('>=', 'Greater than or equal (>=)'), ('<=', 'Smaller than or equal (<=)'), ('regex', 'Regex (removes matches)')], default=None, max_length=10, null=True),
        ),
    ]



# Generated by Django 3.2.13 on 2022-09-23 08:55

import ddm.models.core
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0014_auto_20220922_1640'),
    ]

    operations = [
        migrations.AddField(
            model_name='donationproject',
            name='expected_url_parameters',
            field=models.CharField(blank=True, help_text='Separate multiple parameters with a semikolon (";"). Semikolons are not allowed as part of the expected url parameters.', max_length=500, null=True, verbose_name='Expected URL parameter'),
        ),
        migrations.AddField(
            model_name='donationproject',
            name='url_parameter_enabled',
            field=models.BooleanField(default=False, verbose_name='URL parameter extraction enabled'),
        ),
        migrations.AddField(
            model_name='participant',
            name='extra_data',
            field=models.JSONField(default=ddm.models.core.get_extra_data_default),
        ),
    ]



# Generated by Django 3.2.9 on 2022-05-03 08:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0004_auto_20220502_1810'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationproject',
            name='public_key',
            field=models.BinaryField(),
        ),
    ]



# Generated by Django 3.2.13 on 2022-06-23 19:42

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0007_auto_20220621_1136'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='researchprofile',
            name='api_key',
        ),
        migrations.RemoveField(
            model_name='researchprofile',
            name='api_secret',
        ),
    ]



from django.db.models.signals import post_delete
from django.dispatch import receiver

from ddm.models.core import DonationInstruction, FileUploader


@receiver(post_delete, sender=DonationInstruction)
def reorder_indices(sender, instance, **kwargs):
    # Only execute signal, if model is not deleted due to delete.cascade.
    try:
        queryset = instance.file_uploader.donationinstruction_set.filter(index__gt=instance.index).order_by('index')
        for q in queryset:
            q.index -= 1
            q.save()
    except FileUploader.DoesNotExist:
        pass
    return


@receiver(post_delete, sender=FileUploader)
def reorder_indices(sender, instance, **kwargs):
    # Only execute signal, if model is not deleted due to delete.cascade.
    queryset = FileUploader.objects.filter(project=instance.project, index__gt=instance.index).order_by('index')
    for q in queryset:
        q.index -= 1
        q.save()
    return



# Generated by Django 3.2.13 on 2023-01-10 16:38

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0026_exceptionlogentry_raised_by'),
    ]

    operations = [
        migrations.AddField(
            model_name='exceptionlogentry',
            name='blueprint',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='ddm.donationblueprint'),
        ),
        migrations.AlterField(
            model_name='exceptionlogentry',
            name='exception_type',
            field=models.IntegerField(null=True),
        ),
    ]



# Generated by Django 3.2.13 on 2023-02-27 10:17

import ckeditor_uploader.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0036_alter_processingrule_replacement_value'),
    ]

    operations = [
        migrations.AddField(
            model_name='donationblueprint',
            name='description',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='exp_file_format',
            field=models.CharField(choices=[('json', 'JSON file'), ('csv', 'CSV file')], default='json', max_length=10, verbose_name='Expected file format'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='briefing_text',
            field=ckeditor_uploader.fields.RichTextUploadingField(blank=True, null=True, verbose_name='Briefing Text'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='debriefing_text',
            field=ckeditor_uploader.fields.RichTextUploadingField(blank=True, null=True, verbose_name='Debriefing Text'),
        ),
    ]



# Generated by Django 3.2.13 on 2023-05-25 14:21

import ckeditor_uploader.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0042_auto_20230429_2215'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='questionbase',
            options={'ordering': ['page', 'index']},
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='redirect_target',
            field=models.CharField(blank=True, help_text='Always include <i>http://</i> or <i>https://</i> in the redirect target. If URL parameter extraction is enabled for this project, you can include the extracted URL parameters in the redirect target as follows: "https://redirect.me/?redirectpara=<b>{{participant.data.url_param.URLParameter}}</b>".', max_length=2000, verbose_name='Redirect target'),
        ),
        migrations.AlterField(
            model_name='questionbase',
            name='text',
            field=ckeditor_uploader.fields.RichTextUploadingField(blank=True, help_text='If a question is linked to a File Blueprint, data points from the donated data associated with the linked donation blueprint can be included in the question text. This data can be included as "{{ data }}" in the question text. It is possible to subset the data object (e.g., to include the last datapoint you can use {{ data.0 }} or include advanced rendering options included in the Django templating engine. For a more comprehensive overview and examples see the documentation. Additionally, information directly related to the participant can be included in the question text. This information can be referenced as "{{ participant }}".', null=True),
        ),
    ]



# Generated by Django 3.2.13 on 2023-04-27 15:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0039_alter_donationblueprint_json_extraction_root'),
    ]

    operations = [
        migrations.AddField(
            model_name='participant',
            name='current_step',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



# Generated by Django 3.2.13 on 2022-09-29 11:08

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0017_alter_processingrule_comparison_operator'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='donationblueprint',
            name='extracted_fields',
        ),
        migrations.RemoveField(
            model_name='processingrule',
            name='input_type',
        ),
    ]
